import pytest
import qolab.tableflow as tblfl
import pandas as pd

def test_table_load_noinputs():
    assert tblfl.loadInOutTables() == (None, None)
    assert tblfl.loadInOutTables(inputFileName=None, outputFileName="non_existing_file") == (None, None)

def test_wrong_comment_in_table_file_to_load():
    with pytest.raises(Exception) as exc_info:
        # should raise ParserError
        tblfl.loadInOutTables(inputFileName='tests/tableflow_test_data/tableIn1.csv', outputFileName=None, comment='%')

def test_right_comment_in_table_file_to_load():
    tIn,tOut = tblfl.loadInOutTables(inputFileName='tests/tableflow_test_data/tableIn1.csv', outputFileName=None, comment='#')
    assert type(tIn) == pd.core.frame.DataFrame

def test_table_equality_with_no_output_file_name():
    tIn,tOut = tblfl.loadInOutTables(inputFileName='tests/tableflow_test_data/tableIn1.csv', outputFileName=None, comment='#')
    assert type(tIn) == pd.core.frame.DataFrame
    assert type(tOut) == pd.core.frame.DataFrame
    assert tIn.equals(tOut)
    col0 = tIn.keys()[0]
    vBefore = tIn.at[0, col0]
    tIn.at[0, col0] = vBefore + 1
    assert not tIn.equals(tOut)

def test_table_load_with_in_out_file_names():
    # different filenames, same content for ease of testing
    tIn,tOut = tblfl.loadInOutTables(inputFileName='tests/tableflow_test_data/tableIn1.csv', outputFileName='tests/tableflow_test_data/tableOut1nonProcessed.csv', comment='#')
    assert type(tIn) == pd.core.frame.DataFrame
    assert type(tOut) == pd.core.frame.DataFrame
    assert tIn.equals(tOut)

    # different filenames, different content
    tIn,tOut = tblfl.loadInOutTables(inputFileName='tests/tableflow_test_data/tableIn1.csv', outputFileName='tests/tableflow_test_data/tableOut1pariallyProcessed.csv', comment='#')
    assert type(tIn) == pd.core.frame.DataFrame
    assert type(tOut) == pd.core.frame.DataFrame
    assert not tIn.equals(tOut)
    assert 'out1' in tOut.columns
    assert 'out1' not in tIn.columns

def test_for_existing_row():
    tbl1 = pd.DataFrame( {'a':[1,2,3], 'b':[1,4,6]})
    r = pd.Series({'a':2, 'b':4})
    assert tblfl.ilocRowOrAdd(tbl1, r) == 1

def test_for_existing_row_with_NA():
    # NA in both table and raw should return a hit
    tbl1 = pd.DataFrame( {'a':[1,2,3], 'b':[1,pd.NA,6]})
    r = pd.Series({'a':2, 'b':pd.NA})
    assert tblfl.ilocRowOrAdd(tbl1, r) == 1

    # should insert new row
    tbl1 = pd.DataFrame( {'a':[1,2,3], 'b':[1,4,6]})
    r = pd.Series({'a':2, 'b':pd.NA})
    assert tblfl.ilocRowOrAdd(tbl1, r) == 3

    # should insert new row
    tbl1 = pd.DataFrame( {'a':[1,2,3], 'b':[1,4,6]})
    r = pd.Series({'a':2, 'b':pd.NA})
    assert tblfl.ilocRowOrAdd(tbl1, r) == 3

def test_for_nonexisting_row_and_its_insertion():
    tbl1 = pd.DataFrame( {'a':[1,2,3], 'b':[1,4,6]})
    r = pd.Series({'a':2, 'b':10})
    assert len(tbl1) == 3
    assert tblfl.ilocRowOrAdd(tbl1, r) == 3
    assert len(tbl1) == 4

def test_isRedoNeeded():
    r = pd.Series({'a':2, 'b':4, 'c':pd.NA})
    assert not tblfl.isRedoNeeded(r, ['a','b'])
    assert tblfl.isRedoNeeded(r, ['c'])
    assert tblfl.isRedoNeeded(r, ['non_existing'])
    assert not tblfl.isRedoNeeded(r, ['b', 'c'])

def test_reflowTable():
    tIn,tOut = tblfl.loadInOutTables(inputFileName='tests/tableflow_test_data/tableIn1.csv', outputFileName='tests/tableflow_test_data/tableOut1pariallyProcessed.csv', comment='#')
    tOutRef = tOut.copy()
    # check for warnings
    with pytest.warns(UserWarning):
        tblfl.reflowTable(tIn,tOut)
    
    with pytest.warns(UserWarning):
        tblfl.reflowTable(tIn,tOut,postProcessedColums=['dummyName'])
    
    def frow(row):
        return row
    with pytest.warns(UserWarning):
        tblfl.reflowTable(tIn,tOut, process_row_func=frow)
    
    # now run reflow
    def frow(row, extraInfo=None):
        row['out1'] = row['x']*row['x']
        return row
    assert len(tIn) != len(tOut)
    tblfl.reflowTable(tIn,tOut, process_row_func=frow, postProcessedColums=['out1','out2'])
    assert len(tIn) == len(tOut)
    assert (tOut['out1'] == tOut['x']*tOut['x']).all()
    
    # check that reflow is done
    tOut.loc[tOut['x']==1, 'out1'] = pd.NA
    tblfl.reflowTable(tIn,tOut, process_row_func=frow, postProcessedColums=['out1','out2'])
    assert (tOut['out1'] == tOut['x']*tOut['x']).all()

    # check that reflow is not reprocessed
    tOut.loc[tOut['x']==1, 'out1'] = 12121 # crazy number
    tblfl.reflowTable(tIn,tOut, process_row_func=frow, postProcessedColums=['out1','out2'])
    assert (tOut.loc[tOut['x']==1, 'out1'] == 12121).all() # should not change

    # now we are forcing redo
    tOut.loc[tOut['x']==1, 'out1'] = 12121 # crazy number
    tblfl.reflowTable(tIn,tOut, process_row_func=frow, postProcessedColums=['out1','out2'], redo=True)
    assert not (tOut.loc[tOut['x']==1, 'out1'] == 12121).all() # must not be the same
    assert (tOut['out1'] == tOut['x']*tOut['x']).all()



import justpy as jp
import asyncio
import matplotlib.pyplot as plt

import logging
logger = logging.getLogger('qolab.gui.web')

button_classes = 'bg-blue-500 hover:bg-blue-700 text-white font-bold py-2 px-4 rounded-full'
label_div_classes = 'py-1 px-2'
checkbox_classed  = label_div_classes
labelnames_classes = 'font-bold'
panel_div_classes = 'space-x-4 border'
controls_div_classes = 'flex flex-wrap space-x-4 p-1 border'
controls_group_classes = 'flex flex-wrap space-x-4 p-1'
dict_classes = 'px-2 border-2'
dict_components_classes = 'flex flex-wrap border-2'
input_classes = "m-2 bg-gray-200 border-2 border-gray-200 rounded w-20 text-gray-700 focus:outline-none focus:bg-white focus:border-purple-500"


class QOLPushButton(jp.Button):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.set_classes('bg-blue-500 hover:bg-blue-700 text-white font-bold py-2 px-4 rounded-full')

class QOLPushButtonNoUndo(QOLPushButton):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.set_classes('bg-red-500 hover:bg-red-700 text-white font-bold py-2 px-4 rounded-full')

class QOLParamHeader(jp.Div):
    def __init__(self, label='NameNotSet', value=None, **kwargs):
        super().__init__(**kwargs)
        root = self
        root.label=label
        root.set_classes(label_div_classes)
        root.data[label]=None
        jp.Span(text=f'{label}: ', classes=labelnames_classes, a=root)
        self.setValue(value)

    def getValue(self):
        return self.data[self.label]

    def setValue(self, val):
        # print(f'setting {self.label} to {val}')
        self.data[self.label]=val

class QOLParamReadOnly(QOLParamHeader):
    """ Read Only from the web point of view """
    def __init__(self, label='NameNotSet',  **kwargs):
        super().__init__(label=label, **kwargs)
        root = self
        jp.Span(model=[root, label], a=root)

class QOLParamReadWrite(QOLParamHeader):
    def __init__(self, label='NameNotSet',  **kwargs):
        super().__init__(label=label, **kwargs)
        root = self
        self.input=jp.InputChangeOnly(classes=input_classes, model=[root, label], a=root, spellcheck="false")

class QOLCheckbox(jp.Label):
    def __init__(self, label='NameNotSet', **kwargs):
        super().__init__(**kwargs)
        root = self
        root.data['checked'] = False
        root.label=label
        root.set_classes(checkbox_classed)
        checkbox = jp.Input(type='checkbox', model=[root, 'checked'], classes='form-checkbox', a=root)
        caption = jp.Span(text=label, a=root)

    def getValue(self):
        return self.data['checked']
    
    def setValue(self, val):
        self.data['checked'] = val

class QOLTimeLog(jp.Div):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        root = self
        root.traces = None; # log must have 'plot' and 'clear_data' methods
        root.set_classes(panel_div_classes)
        dcontrols = jp.Div(a=root)
        dcontrols.set_classes(controls_div_classes)
        bclear = QOLPushButtonNoUndo(a=dcontrols, text='Clear log', click=self._clear_data)
        breplot = QOLPushButton(a=dcontrols, text='Replot', click=self._replot)
        self.update_interval = QOLParamReadOnly(label='UpdateInterval',value=5,a=dcontrols)
        self.save_controls = QOLSaveControls(component_with_data=self.traces, a=dcontrols)

        self.chart = jp.Matplotlib(a=root)
        self.start_task()

    def setTraces(self,component_with_data=None):
        self.traces = component_with_data
        self.save_controls.component_with_data=component_with_data

    def start_task(self):
        jp.run_task( self.update_loop() )

    def stop_tasks(self):
        self.save_controls.stop_tasks()
        pass

    async def _clear_data(self, msg):
        self.clear_data()
        await self.update()

    def clear_data(self):
        traces = self.traces
        if traces is not None:
            traces.clear_data()
            self.plot()

    async def _replot(self, msg):
        self.plot()
        await self.update()

    async def update_loop(self, update_interval=None):
        if update_interval is not None:
            self.update_interval.setValue(update_interval)
        while True:
            self.plot()
            try:
                strwtime = self.update_interval.getValue()
                wtime = float(strwtime)
            except ValueError:
                wtime =2
                print(f'Warning cannot convert {strwtime} setting to {wtime}')
            await self.update()
            await asyncio.sleep(wtime)

    def plot(self):
        traces = self.traces
        f = plt.figure(figsize=(8,6), tight_layout=True)
        if traces is None:
            plt.title('Log data is unavailable')
        else:
            traces.plot()
        self.chart.set_figure(f)
        plt.close(f)


class QOLSaveControls(jp.Div):
    def __init__(self, component_with_data=None, **kwargs):
        super().__init__(**kwargs)
        root = self
        root.set_classes(controls_group_classes)

        root.component_with_data = component_with_data
        root.getNextDataFile = None
        root.bsave=QOLPushButton(a=root, text='Save', name='Save', click=self._save)
        root.bnext_file=QOLPushButton(a=root, text='Next file', name='NextFile', click=self._next_file)
        self.autosave_flag=QOLCheckbox(label='autosave', a=root )
        self.autosave_interval = QOLParamReadOnly(label='AutoSaveInterval', value=10, a=root)
        self.file_name=QOLParamReadOnly(label='FileName', a=root)
        self.start_task()

    def start_task(self):
        jp.run_task( self.autosave_loop() )

    def stop_tasks(self):
        pass

    async def autosave_loop(self):
        while True:
            if self.autosave_flag.getValue():
                self.save()
            await asyncio.sleep(10)

    async def _save(self, msg):
        self.save()

    def save(self):
        fname = self.file_name.getValue()
        if fname is None:
            return
        if self.component_with_data is None:
            return
        logger.debug(f'saving to {fname}')
        self.component_with_data.save(fname)

    async def _next_file(self, msg):
        self.next_file()
        await self.update()

    def next_file(self):
        if self.getNextDataFile is not None:
            fname = self.getNextDataFile()
            logger.info(f'Data will be saved to {fname}')
            self.file_name.setValue( fname )
            if self.component_with_data is not None:
                self.component_with_data.clear_last_saved_pos()

class QOLDictionary(jp.Div):
    def __init__(self, container=None, **kwargs):
        super().__init__(**kwargs)
        root = self
        root.set_classes(dict_classes)
        self.dlabel=jp.Div(a=root, classes='bg-gray-100')
        self.dlabel.on('click', self._toggle_show)
        root.slabel=jp.Span(text = self.name, a=root.dlabel)
        root.slabel.set_classes(labelnames_classes)
        self.c=jp.Div(a=root)
        root.c.set_classes(dict_components_classes)
        root.container=container
        self.display_container_dictionary()

    async def _toggle_show(self, msg):
        self.c.show = not self.c.show

    def display_container_dictionary(self):
        self.c.delete_components()
        if self.container is None:
            return
        for k,v in self.container.items():
            if not isinstance(v, dict):
                QOLParamReadOnly(label=k, value=v, a=self.c)
            else:
                QOLDictionary(container=v, name=k, a=self.c)



def gui_test():
    return wp

if __name__ == '__main__': 
    wp = jp.WebPage(delete_flag=False)
    rw = QOLParamReadWrite(label='ReadWriteParam', a=wp)
    rw.setValue(12345)

    log = QOLTimeLog(a=wp)
    log.save_controls.getNextDataFile=lambda: "data.dat"

    def test(self, msg):
        print(rw.getValue())
    QOLPushButtonNoUndo(text='Danger', a=wp, onclick=test)
    # sc = QOLSaveControls(a=wp)
    d={}
    d['a']='astring'
    d['n']=7
    d['d']={}
    for i in range(20):
        d['d'][i]=i
    QOLDictionary(a=wp, name='d dictionary',  container=d)


    jp.justpy(gui_test)




from qolab.hardware.scpi import SCPIinstr
from qolab.hardware.basic import BasicInstrument

class Multimeter(BasicInstrument):
    def __init__(self, *args, **kwds):
        BasicInstrument.__init__(self, *args, **kwds)
        self.config['Device type']='Multimeter'
        self.config['Device model'] = 'Generic Multimeter generator Without Hardware interface'
        self.config['FnamePrefix'] = 'Multimeter'
        self.config['Device model'] = 'Generic Multimeter Without Hardware interface'
        self.config['FnamePrefix'] = 'multimeter'
        self.deviceProperties.update({})

    # Minimal set of methods to be implemented.
    def getVdc(self):
        """ Report DC Voltage """
        print('getVdc is not implemented')
        return None
    def getVac(self):
        """ Report AC Voltage """
        print('getVac is not implemented')
        return None
    def getAdc(self):
        """ Report DC Current """
        print('getAdc is not implemented')
        return None
    def getAac(self):
        """ Report AC Current """
        print('getAac is not implemented')
        return None
    def getResistance(self):
        """ Report Resistance """
        print('getResistance is not implemented')
        return None
    def getResistance4Wires(self):
        """ Report Resistance with 4 wire method """
        print('getResistance4Wires is not implemented')
        return None
    def getDiode(self):
        """ Report Diode Voltage drop """
        print('getDiode is not implemented')
        return None
    def getFreq(self):
        """ Report Frequency """
        print('getFreq is not implemented')
        return None


class MultimeterSCPI(SCPIinstr, Multimeter):
    """     
    Do not instantiate directly, use
    rm = pyvisa.ResourceManager()
    MultimeterSCPI(rm.open_resource('TCPIP::192.168.0.2::INSTR'))
    """
    pass
    def __init__(self, resource, *args, **kwds):
        SCPIinstr.__init__(self, resource)
        Multimeter.__init__(self, *args, **kwds)
        self.config['DeviceId'] = str.strip(self.idn)

from .hp_34401 import HP_34401
from .bk_5491 import BK_5491






from qolab.hardware.scpi import SCPIinstr
from qolab.hardware.basic import BasicInstrument

class Lockin(BasicInstrument):
    def __init__(self, *args, **kwds):
        BasicInstrument.__init__(self, *args, **kwds)
        self.config['Device type']='Lockin'
        self.config['FnamePrefix'] = 'lockin'
        self.config['Device model'] = 'Generic Lockin Without Hardware interface'
        self.config['FnamePrefix'] = 'lockin'
        self.deviceProperties.update({'FreqInt', 'FreqExt', 'Harm', 'SinAmpl', 'SinOffset',
                'RefPhase',
                'Sensitivity', 'TimeConstan', 'FilterSlope', 'EquivalentNoiseBW'})
    # Minimal set of methods to be implemented.
    pass

class LockinSCPI(SCPIinstr, Lockin):
    """     
    Do not instantiate directly, use
    rm = pyvisa.ResourceManager()
    LockinSCPI(rm.open_resource('TCPIP::192.168.0.2::INSTR'))
    """
    pass
    def __init__(self, resource, *args, **kwds):
        SCPIinstr.__init__(self, resource)
        Lockin.__init__(self, *args, **kwds)
        self.config['DeviceId'] = str.strip(self.idn)

from .srs_sr865a import SRS_SR865A






import numpy as np
from scipy.fft import fft, rfft, irfft, rfftfreq

def spectrum(t, y):
    """
    Spectrum of real value signal with stripped away zero frequency component.
    Preserves amplitude of a coherent signal (Asin(f*t)) independent of sampling
    rate and time interval.
    """
    N = t.size
    dt=np.mean(np.diff(t))
    freq=rfftfreq(N, dt)
    # y= y - np.mean(y)
    yf = rfft(y)
    yf *= 2/N; # produce normalized amplitudes
    return freq[1:], yf[1:]; # strip of boring freq=0

def noise_density_spectrum(t,y):
    """
    Calculates noise amplitude spectral density (ASD), the end results has unitis of y/sqrt(Hz)
    i.e. it does sqrt(PSD) where PSD is powerd spectrum density.
    Preserves the density independent of sampling rate and time interval.
    """
    freq, yf = spectrum(t, y)
    yf = yf*np.sqrt(t[-1]-t[0]) # scales with 1/sqrt(RBW)
    return freq, yf

def noise_spectrum_smooth(fr, Ampl, Nbins=100):
    """
    Smooth spectrum, especially at high frequency end.
    Could be thought as logarithmic spacing running average.
    Since we assume the spectrum of the nose, we do power average (rmsq on amplitudes)
    Assumes that set of frequencies is positive and equidistant set.
    Also assumes that frequencies do not contain 0.
    """

    frEdges = np.logspace(np.log10(fr[0]), np.log10(fr[-1]), Nbins)
    frCenter = np.zeros(frEdges.size-1)
    power = np.zeros(frEdges.size-1)
    for i, (frStart, frEnd) in enumerate(zip(frEdges[0:-1], frEdges[1:])):
        # print (f"{i=} {frStart=} {frEnd}")
        ind = (frStart <= fr) & (fr <= frEnd)
        frCenter[i] = np.mean( fr[ind] ) 
        power [i] = np.mean( np.power( np.abs(Ampl[ind]),2) )
    ind = np.logical_not(np.isnan(frCenter))
    frCenter = frCenter[ind]
    power = power[ind]
    # print(f'{frCenter=} {power=}')
    return frCenter,  np.sqrt( power )




from qolab.hardware.basic import BasicInstrument
from qolab.hardware.daq import DAQ
import ue9

class LabJackUE9(DAQ):
    """
    This custom modification of the stock LabJack UE9 class provided by LabJack as ue9
    """

    def __init__(self, *args, debug = False, autoOpen = True, ethernet=False, ipAddress=None, **kargs):
        """
        LabJack UE9 can be contacted via TCP, use
        LabJackUE9(ethernet=True, ipAddress="192.168.1.209")
        """
        super().__init__(*args, **kargs)
        self.config['Device model'] = 'LabJack UE9'
        self.AnalogInputsNum=4
        self.AnalogOutputsNum=2

        self.daq = ue9.UE9(debug=debug, autoOpen=autoOpen, ethernet=ethernet, ipAddress=ipAddress)

        # For applying the proper calibration to readings.
        c=self.daq.getCalibrationData()

        # by evmik
        # fixing missing slope for gain '0'
        c['AINSlopes']['0']= 0.0000775030

    def getAIN(self, chNum):
        """ get analog input (AIN) voltage """
        BipGain = 8
        Resolution = 12
        SettlingTime = 0
        # BipGain = 8 -> bipolar range (-5V, +5V) gain 1
        # UE9 default BipGain = 0 -> signal range (0V, +5V) gain 1
        # other BipGain could be:
        #   0 = Unipolar Gain 1, 1 = Unipolar Gain 2,
        #   2 = Unipolar Gain 4, 3 = Unipolar Gain 8,
        #   8 = Bipolar Gain 1
        return self.daq.getAIN(chNum, BipGain = BipGain, Resolution = Resolution, SettlingTime = SettlingTime)

    def setDAC(self, chNum, volts):
        """ set digital to analog (DAC) output voltage """
        if chNum == None or volts == None:
            print("setOutputCh needs chNum and volts to be set")
            return 0
        bits = self.daq.voltageToDACBits(volts, dacNumber = chNum)
        # out is completely bogus for DAC settings in UE9
        out=self.daq.singleIO(IOType=5, Channel=chNum, DAC=bits)
        return volts

    def close(self):
        self.daq.close()



if __name__ == '__main__': 
    daq = LabJackUE9()
    print("testing")
    print('------ Header start -------------')
    print(str.join('\n', daq.getHeader()))
    print('------ Header ends  -------------')



from qolab.hardware.basic import BasicInstrument

class DAQ(BasicInstrument):
    # Minimal set of methods to be implemented by a RFGenerator
    def __init__(self, *args, **kwds):
        BasicInstrument.__init__(self, *args, **kwds)
        self.config['Device type']='DAQ'
        self.config['Device model'] = 'Generic DAQ Without Hardware interface'
        self.config['FnamePrefix'] = 'daq'
        self.deviceProperties.update({'AnalogInputsNum', 'AnalogOutputsNum'})
        
        # this is device dependent
        self.AnalogInputsNum=0
        self.AnalogOutputsNum=0

from .labjack_ue9 import LabJackUE9




"""
Provide basic method to process data describing tables
Created by Eugeniy E. Mikhailov 2024/05/27

The basic idea that we will have an *input* table
with data description and we (re)generate *output* table
based on the input table with processed rows.

If output table already have processed rows with entries different from NA
such rows are skipped.

Super handy for bulk processing data files where only a few parameters changed.
"""

import pandas as pd
import warnings

def loadInOutTables(inputFileName=None, outputFileName=None, comment=None):
    if not inputFileName:
        return None, None

    if not comment:
        comment = '#'

    tIn = pd.read_csv(inputFileName, comment=comment)
    tIn.columns = tIn.columns.str.removeprefix(' '); # clean up leading white space in columns names

    try:
        tOut=pd.read_csv(outputFileName, comment=comment)
    except Exception:
        tOut=tIn.copy(deep=True)

    return tIn, tOut

def ilocRowOrAdd(tbl, row):
    # Find similar 'row' in 'tbl', NA in both set treated as a hit.
    # if similar row not found, insert it.
    tSub = tbl[row.keys()]
    res = (tSub == row) | (tSub.isna() & row.isna() )
    res = res.all(axis=1) # which rows coincide
    if res.any():
        # we have a similar row
        i = res[res].index[0]
    else:
        # we need to create new row since tbl does not has it
        i=len(tbl)
        updateTblRowAt(tbl, i, row)
    return i

def updateTblRowAt(tbl, i, row):
    for k in row.keys():
        tbl.at[i, k] = row[k]
    return

def isRedoNeeded(row, cols2check):
    # redo is required if *all* required entries in cols2check are NA
    # or we are missing columns in cols2check list
    for c in cols2check:
        if c not in row.keys():
            return True
    if row[cols2check].isna().all():
        return True
    return False

def reflowTable(tIn, tOut, process_row_func=None, postProcessedColums=None, extraInfo=None, redo=False):
    # update tOut in place based on the inputs specified in tIn
    # effectively maps unprocess rows in to process_row_func
    # - postProcessedColums is a list of column names which need to be generated
    # - extraInfo is dictionary of additional parameter supplied to process_row_func
    # - process_row_func expected to behave like:
    #   rowOut = process_row_func(rowIn, extraInfo=userInfo)
    # - redo controls if reflow is needed unconditionally (i.e. force reflow)
    if not process_row_func:
        warnings.warn("process_row_func is not provided, exiting reflowTable")
        return
    if not postProcessedColums:
        warnings.warn("postProcessedColums are not provided, exiting reflowTable")
        return

    for index, rowIn in tIn.iterrows():
        iOut = ilocRowOrAdd(tOut, rowIn)
        rowOutBefore = tOut.iloc[iOut]

        if not (redo or isRedoNeeded(rowOutBefore, postProcessedColums) ):
            continue

        # processing data describing row
        rowOut = process_row_func(rowOutBefore, extraInfo=extraInfo)
        updateTblRowAt(tOut, iOut, rowOut)




from qolab.hardware.scpi import SCPIinstr
from qolab.hardware.basic import BasicInstrument

class PowerSupply(BasicInstrument):
    # Minimal set of methods to be implemented by a Power Supply
    def __init__(self, *args, **kwds):
        BasicInstrument.__init__(self, *args, **kwds)
        self.config['Device type']='PowerSupply'
        self.config['Device model'] = 'Generic Power Supply generator Without Hardware interface'
        self.config['FnamePrefix'] = 'power_supply'
        self.deviceProperties.update({})

class PowerSupplySCPI(SCPIinstr, PowerSupply):
    """     
    Do not instantiate directly, use
    rm = pyvisa.ResourceManager()
    PowerSupplySCPI(rm.open_resource('TCPIP::192.168.0.2::INSTR'))
    or 
    PowerSupplySCPI(rm.open_resource('USB0::10893::4354::MY61001869::0::INSTR'))
    """
    def __init__(self, resource, *args, **kwds):
        SCPIinstr.__init__(self, resource)
        PowerSupply.__init__(self, *args, **kwds)
        self.config['DeviceId'] = str.strip(self.idn)

from .keysight_e3612a import KeysightE3612A




import logging
# this should be done before justpy is called or log formatter does not work
logging.basicConfig(format='%(asctime)s %(levelname)8s %(name)s: %(message)s', datefmt='%m/%d/%Y %H:%M:%S')

# import justpy as jp
import pyvisa
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt

import qolab.tsdb as tsdb
import asyncio
import time

# import qolab.gui.web as gui
from qolab.hardware import BasicInstrument
from qolab.hardware.scope import SDS1104X
from qolab.hardware.rf_generator import AgilentE8257D
from qolab.hardware.lockin import SRS_SR865A
from qolab.feedback import PID
from qolab.data import TraceSetSameX, TraceXY, Trace
from qolab.hardware.i_server import I800

l = logging.getLogger('qolab.gui.web')
l.setLevel(logging.INFO)
l = logging.getLogger('qolab.tsdb')
l.setLevel(logging.INFO)
logger = logging.getLogger('Magnetometer')
logger.setLevel(logging.INFO)

def getConfig(apparatus):
    config = apparatus.config.copy()
    ai = apparatus.instruments
    for n, i in ai.items():
        config[n]=i.getConfig()
    return config

class Apparatus(BasicInstrument):
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        self.setLockinSlope(0)

    @BasicInstrument.tsdb_append
    def getBCurrent(self):
        scope = self.instruments['scope.monitor']
        ch=scope.getTrace(2, decimate=False)
        bCurrent = np.mean(ch.y.values); # mV -> mA
        return bCurrent

    def getErrAndEit(self):
        scope = self.instruments['scope.feedback']
        # we need to freeze scope since this two values are taken on the same scope
        trg_mode = scope.getTriggerMode()
        scope.setTriggerMode('STOP')
        err = self.getErr()
        eit = self.getEIT()
        scope.setTriggerMode(trg_mode)
        return(err,eit)



    @BasicInstrument.tsdb_append
    def getErr(self):
        scope = self.instruments['scope.feedback']
        ch1=scope.getTrace(1, decimate=False)
        err = np.mean(ch1.y.values)
        return err

    @BasicInstrument.tsdb_append
    def getEIT(self):
        scope = self.instruments['scope.feedback']
        ch3=scope.getTrace(3, decimate=False)
        eit = np.mean(ch3.y.values)
        return eit

    @BasicInstrument.tsdb_append
    def setFeefback(self, val):
        freq0 = self.getRFreq()
        self.setRFreq(freq0+val)

    @BasicInstrument.tsdb_append
    def getRFreq(self):
        rfgen = self.instruments['rfgen']
        return rfgen.getFreqFixed()

    @BasicInstrument.tsdb_append
    def setRFreq(self,val):
        rfgen = self.instruments['rfgen']
        rfgen.setFreqFixed(val)

    @BasicInstrument.tsdb_append
    def getRFAmplitude(self):
        rfgen = self.instruments['rfgen']
        return rfgen.getRFAmplitude()

    @BasicInstrument.tsdb_append
    def setRFAmplitude(self, val):
        rfgen = self.instruments['rfgen']
        return rfgen.setRFAmplitude(val)

    @BasicInstrument.tsdb_append
    def setLockinSlope(self, val):
        self.lockin_slope=val

if __name__ == '__main__': 

    # TSDB logger setting
    tsdb_client = tsdb.Client('influx', 'http://lumus.physics.wm.edu:8428', database='qolab')
    tsdb_ingester = tsdb.Ingester(tsdb_client, batch=11, measurement_prefix='VAMPIRE.HighPower')

    # creating Apparatus with all instruments to be logged
    apparatus = Apparatus(tsdb_ingester=tsdb_ingester, device_nickname='magnetometer',)
    apparatus.config['Device type'] = 'QOL VAMPIRE HighPower magnetometer'
    apparatus.config['Device model'] = 'v0.1'
    apparatus.config['FnamePrefix'] = 'magnetometer_eit'
    apparatus.config['SavePath'] = '/mnt/qol_grp_data/data.VAMPIRE.HighPower'

    app_nickname=apparatus.config['DeviceNickname']

    logger.info("Accessing hardware")
    rm = pyvisa.ResourceManager()
    instr_scope=rm.open_resource('TCPIP::192.168.0.61::INSTR')
    scope_fdbk = SDS1104X(instr_scope, device_nickname='.'.join([app_nickname, 'scope.feedback']), tsdb_ingester=tsdb_ingester)
    instr_scope=rm.open_resource('TCPIP::192.168.0.62::INSTR')
    scope_mon = SDS1104X(instr_scope, device_nickname='.'.join([app_nickname, 'scope.monitor']), tsdb_ingester=tsdb_ingester)
    instr_rfgen=rm.open_resource('TCPIP::192.168.0.114::INSTR')
    rfgen = AgilentE8257D(instr_rfgen, device_nickname='.'.join([app_nickname, 'rfgen']), tsdb_ingester=tsdb_ingester)
    instr_lockin=rm.open_resource('TCPIP::192.168.0.51::INSTR')
    lockin = SRS_SR865A(instr_lockin, device_nickname='.'.join([app_nickname, 'lockin']), tsdb_ingester=tsdb_ingester)

    # adding instruments to apparatus
    apparatus.instruments={}
    ai = apparatus.instruments
    ai['rfgen'] = rfgen
    ai['lockin']   = lockin
    ai['cellTemperatureController'] = I800(device_nickname='.'.join([app_nickname, 'cellTemperatureController']), tsdb_ingester=tsdb_ingester)
    # Do not add scope.feedback to apparatus.instruments it will create recursive loop
    # in the saved data file
    # ai['scope.feedback'] = scope_fdbk
    scope_fdbk.config['SavePath'] = '/mnt/qol_grp_data/data.VAMPIRE.HighPower'
    scope_fdbk.config['FnamePrefix'] = 'scope_eit'
    # scope_fdbk.setRoll(False)
    # scope_fdbk.setRun(True)
    # scope_fdbk.setTimePerDiv(0.0005)
    # scope_fdbk.setChanVoltsPerDiv(1, 0.02)
    # scope_fdbk.setChanVoltsPerDiv(3, 0.02)
    ai['scope.monitor'] = scope_mon
    # scope_mon.setRoll(False)
    # scope_mon.setRun(True)
    # scope_mon.setTimePerDiv(0.0005)
    # scope_mon.setChanVoltsPerDiv(2, 0.002)
    # scope_mon.setChanVoltageOffset(2, -0.0726)

    scope_fdbk.config['tags'] = {}

    cfFreq = [6_833_980_000, 6_834_686_000, 6_835_393_000]
    m      = [-2, 0, 2]
    #rfgen.setSweepCentralFreq(6_833_980_000)
    rfgen.setSweepCentralFreq(6_834_686_000)
    #rfgen.setSweepCentralFreq(6_835_393_000)

    for fr,m in zip(cfFreq, m):
        rfgen.setSweepCentralFreq(fr)
        print(f'time to settle for {m=} with {fr=}')
        time.sleep(10)
        print('settling done')

        # if we want to save new set of traces repeat this two commands
        scope_fdbk.config['tags']['apparatus']=getConfig(apparatus)
        scope_fdbk.save( maxRequiredPoints=1000 )

    tsdb_ingester.commit()





from qolab.hardware.basic import BasicInstrument
from qolab.hardware.rf_generator import RFGeneratorSCPI
from qolab.data.trace import Trace
import re
import numpy as np

class AgilentE8257D(RFGeneratorSCPI):
    """ Agilent E8257D RF generator """
    def __init__(self, resource, *args, **kwds):
        super().__init__(resource, *args, **kwds)
        self.resource.read_termination='\n'
        self.config['Device model'] = 'Agilent E8257D'
        self.deviceProperties.update({
            'RFPowerState',
            'RFAmplitude',
            'ModulationState',
            'FM1State',
            'FM1Source',
            'FM1ModulationDepth',
            'FM2State',
            'FM2Source',
            'FM2ModulationDepth',
            'FrequencyMode', # sweep or continious
            'SweepCentralFreq', 'SweepSpan',
            })

    """ Note: Fixed frequency and Center frequency (of sweep)  are different in this model """

    @BasicInstrument.tsdb_append
    def getFreqFixed(self):
        qstr = f':FREQuency:Fixed?'
        rstr = self.query(qstr)
        return( float(rstr) )

    @BasicInstrument.tsdb_append
    def setFreqFixed(self, freq):
        cstr = f':FREQuency:FIXED {freq}Hz'
        self.write(cstr)

    @BasicInstrument.tsdb_append
    def getSweepCentralFreq(self):
        qstr = f':FREQuency:CENTer?'
        rstr = self.query(qstr)
        return( float(rstr) )

    @BasicInstrument.tsdb_append
    def setSweepCentralFreq(self, cfreq):
        cstr = f':FREQuency:CENTer {cfreq}Hz'
        self.write(cstr)

    @BasicInstrument.tsdb_append
    def getSweepSpan(self):
        qstr = f':FREQuency:SPAN?'
        rstr = self.query(qstr)
        return( float(rstr) )

    @BasicInstrument.tsdb_append
    def setSweepSpan(self, span):
        cstr = f':FREQuency:SPAN {span}Hz'
        self.write(cstr)

    @BasicInstrument.tsdb_append
    def setSweep(self, cfreq=None, span=None):
        if cfreq is not None:
            self.setSweepCentralFreq(cfreq)
        if cfreq is not None:
            self.setSweepSpan(span)

    @BasicInstrument.tsdb_append
    def getModulationState(self):
        return(int(self.query(f':MODulation:STATe?')))

    @BasicInstrument.tsdb_append
    def setModulationState(self, val):
        rstr = self.write(f':MODulation:STATe {val}')

    @BasicInstrument.tsdb_append
    def getRFPowerState(self):
        return(int(self.query(f':OUTPut:STATe?')))

    @BasicInstrument.tsdb_append
    def setRFPowerState(self, val):
        rstr = self.write(f':OUTPut:STATe {val}')

    @BasicInstrument.tsdb_append
    def getRFAmplitude(self):
        return(float(self.query(f':POWer:AMPLitude?')))

    @BasicInstrument.tsdb_append
    def setRFAmplitude(self, val):
        rstr = self.write(f':POWer:AMPLitude {val}')

    @BasicInstrument.tsdb_append
    def getFM1ModulationDepth(self):
        return(float(self.query(f':FM1:Deviation?')))

    @BasicInstrument.tsdb_append
    def setFM1ModulationDepth(self, val):
        rstr = self.write(f':FM1:Deviation {val}')

    @BasicInstrument.tsdb_append
    def getFM2ModulationDepth(self):
        return(float(self.query(f':FM2:Deviation?')))

    @BasicInstrument.tsdb_append
    def setFM2ModulationDepth(self, val):
        rstr = self.write(f':FM2:Deviation {val}')

    @BasicInstrument.tsdb_append
    def getFM1Source(self):
        return(str(self.query(f':FM1:Source?')))

    @BasicInstrument.tsdb_append
    def setFM1Source(self, val):
        rstr = self.write(f':FM1:Source {val}')

    @BasicInstrument.tsdb_append
    def getFM2Source(self):
        return(str(self.query(f':FM2:Source?')))

    @BasicInstrument.tsdb_append
    def setFM2Source(self, val):
        rstr = self.write(f':FM2:Source {val}')

    @BasicInstrument.tsdb_append
    def getFM1State(self):
        return(int(self.query(f':FM1:State?')))

    @BasicInstrument.tsdb_append
    def setFM1State(self, val):
        rstr = self.write(f':FM1:State {val}')

    @BasicInstrument.tsdb_append
    def getFM2State(self):
        return(int(self.query(f':FM2:State?')))

    @BasicInstrument.tsdb_append
    def setFM2State(self, val):
        rstr = self.write(f':FM2:State {val}')

    @BasicInstrument.tsdb_append
    def getFrequencyMode(self):
        return(str(self.query(f':Frequency:Mode?')))

    @BasicInstrument.tsdb_append
    def setFrequencyMode(self, val):
        rstr = self.write(f':Frequency:Mode {val}')

    @BasicInstrument.tsdb_append
    def startFrequencySweep(self):
        self.setFrequencyMode('Sweep')

    def stopFrequencySweep(self):
        self.setFrequencyMode('FIXED')

if __name__ == '__main__': 
    import pyvisa
    print("testing")
    rm = pyvisa.ResourceManager()
    print(rm.list_resources())    
    instr=rm.open_resource('TCPIP::192.168.0.114::INSTR')
    rfgen = AgilentE8257D(instr)
    print('------ Header start -------------')
    print(str.join('\n', rfgen.getHeader()))
    print('------ Header ends  -------------')




import platform
import re
import os
from datetime import date

def filename2os_fname( fname ):
    # filename2os_fname translate Win or Linux fname to OS dependent style
    # takes in account the notion of 'Z:' drive on different systems
    # Z:\dir1\dir2\file   <==> /mnt/qol_grp_data/dir1/dir2/file
    if platform.system() == 'Windows':
        fname = re.sub('/mnt/qol_grp_data', 'Z:', fname)
    else:
        fname = re.sub('Z:', '/mnt/qol_grp_data', fname)
        fname = re.sub(r'\\', '/', fname)

    fname = os.path.normpath(fname)
    return (fname)


def get_runnum(data_dir):
    # For the provided data_dir:
    #  reads, increments data counter and saves it back.
    # If necessary creates counter file and full path to it.
    # example
    #  get_runnum('Z:\Ramsi_EIT\data\')
    #  get_runnum('/mnt/qol_grp_data/data')
    data_dir = filename2os_fname( data_dir );
    if not os.path.exists(data_dir):
            os.mkdir(data_dir)
    if not os.path.isdir(data_dir):
        print(f"ERROR: cannot create directory for data: {data_dir}")
        print("Will use current dir for storage")
        data_dir = "."

    runnumpath = os.path.join(data_dir, 'autofile')
    # convert to OS dependent way
    runnumpath = filename2os_fname( runnumpath );

    if not os.path.exists(runnumpath):
            os.mkdir(runnumpath)
    runnum_file = os.path.join(runnumpath, 'runnum.dat');
    runnum_file = filename2os_fname( runnum_file );

    run_number = 0
    if os.path.isfile(runnum_file):
        with open(runnum_file, 'r') as f:
            content = f.readlines()
            run_number = int(content[0])
            f.close()

    # Increment it and fold if needed
    run_number = run_number + 1;
    # Important: we are using five digit counters to synchronize
    # with qol_get_next_data_file.m
    if run_number > 99999:
        run_number = 0

    with open(runnum_file, 'w') as f:
        f.write(f'{run_number}')
        f.close()
    return(run_number)

def get_next_data_file(prefix, savepath, run_number=None, datestr=None, date_format='%Y%m%d', extension='dat'):
    if run_number is None:
        run_number = get_runnum( savepath )
    today = date.today()
    if datestr is None:
        datestr = today.strftime(date_format)
    fname = os.path.join(savepath, f'{prefix}_{datestr}_{run_number:05d}.{extension}')
    return(fname)

def save_table_with_header(fname, data, header='', comment_symbol='%', skip_headers_if_file_exist=False, item_format='e', item_separator='\t', compressionmethod=None, compresslevel=9):
    # itemFormat examples: 'e', '.15e', 'f'
    # `compressionmethod` can be
    #   None - no compression
    #   "gzip" - gzip method of compression
    # `compresslevel`: 9 the highest compression, 0 no compression at all, as it is defined for gzip in Lib/gzip.py
    fname = filename2os_fname(fname)
    file_exist_flag = os.path.exists(fname)
    item_format=str.join('', ['{', f':{item_format}', '}'])
    _open = open # standard file handler
    if compressionmethod == 'gzip':
        import gzip
        _open = lambda fname, mode: gzip.open( fname, mode, compresslevel = compresslevel)
    with _open(fname, 'ab') as f:
        if not (file_exist_flag and skip_headers_if_file_exist):
            for l in header:
                f.write(f'{comment_symbol} {l}\n'.encode('utf-8'))
        if data is not None:
            for r in data:
                l=item_separator.join( map(item_format.format, r))
                f.write(l.encode('utf-8'))
                f.write('\n'.encode('utf-8'))
        f.close()
    return(fname)




import logging
# this should be done before justpy is called or log formatter does not work
logging.basicConfig(format='%(asctime)s %(levelname)8s %(name)s: %(message)s', datefmt='%m/%d/%Y %H:%M:%S')

import pyvisa
import numpy as np
import platform
import time
from tqdm import tqdm as pbar
import matplotlib.pyplot as plt


from qolab.hardware import BasicInstrument
from qolab.data import TraceSetSameX, TraceXY, Trace
from qolab.hardware.rf_generator import QOL_LMX2487
from qolab.hardware.daq import LabJackUE9
from qolab.hardware.power_supply.keysight_e3612a import KeysightE3612A
import qolab.tsdb as tsdb

l = logging.getLogger('qolab.tsdb')
l.setLevel(logging.INFO)
logger = logging.getLogger('BfieldDriver')
logger.setLevel(logging.INFO)

class BfieldDriver(KeysightE3612A):
    """ need to set power supply """
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        self.config['Device type'] = 'B field coil driver based on Keysight E3612A power supply'
        self.config['Device model'] = 'v0.1'
        self.deviceProperties = self.deviceProperties.union({'B', 'Bslope_TperA', 'CoilAssignment'})
        """"
        Rough magnetic field calibration of the 3 axes coils in large magnetic shield
        - Ch1: 70mA -> 650 kHz shift for delta m = 2
        - Ch2: 70mA -> 700 kHz shift for delta m = 2
        - Ch2: 70mA -> 659 kHz shift for delta m = 2
        A better calibration obtained on 20220601 see file 
          20220601.magnetic_field_callibration/calibration_currentToB.dat
        """
        # B response to current in a given channel
        # Olivia's Calibration with flux gate magnetometer used from 2023/04/28
        #  +/-     0.0000007
        self._Bslope_TperA = {
                1: 0.00065859269,
                2: 0.00070732580,
                3: 0.00066754994
                }
        """
        # Irina's Calibration used prior 2023/04/27
        self._Bslope_TperA = {
                1: 0.0006571429,
                2: 0.0007085714,
                3: 0.0006675714
                }
        # Eugeniy's Calibration
        self._Bslope_TperA = {
                1: 0.0006574710928926532,
                2: 0.0007064314754023079,
                3: 0.0006635058865577695
                }
        """
        # assuming that Ch1 controls Bz, Ch2 -> Bx, Ch3 -> By
        self._coil_assignment = { 'chX': 2, 'chY': 3, 'chZ': 1 }

    def getBslope_TperA(self):
        return self._Bslope_TperA

    def getCoilAssignment(self):
        return self._coil_assignment

    def getB(self):
        Bslope_TperA = self.getBslope_TperA()
        coil_assignment = self.getCoilAssignment()
        chX = coil_assignment['chX']
        chY = coil_assignment['chY']
        chZ = coil_assignment['chZ']

        Ix = self.getChanIout(chX)
        Iy = self.getChanIout(chY)
        Iz = self.getChanIout(chZ)
        Bx = Ix * Bslope_TperA[chX]
        By = Iy * Bslope_TperA[chY]
        Bz = Iz * Bslope_TperA[chZ]
        Bmag = float(np.sqrt(Bx*Bx + By*By + Bz*Bz))
        theta = float(np.arccos(Bz/Bmag))
        phi  =  float(np.arctan2(By, Bx))
        theta_degree = float(theta/np.pi*180)
        phi_degree = float(phi/np.pi*180)
        return { 'Bmag': Bmag,
                'theta': theta, 'phi': phi,
                'theta_degree': theta_degree, 'phi_degree': phi_degree,
                'Bx': float(Bx),
                'By': float(By),
                'Bz': float(Bz),
                }

    def setB(self, Bmag=50e-6, theta=0, phi=0):
        """ Sets B field currents based on B (in T) and angles theta, and phi """
        self._Bmag = Bmag
        self._theta = theta
        self._phi = phi

        Bx = Bmag*np.sin(theta)*np.cos(phi)
        By = Bmag*np.sin(theta)*np.sin(phi)
        Bz = Bmag*np.cos(theta)

        Bslope_TperA = self.getBslope_TperA()
        coil_assignment = self.getCoilAssignment()
        chX = coil_assignment['chX']
        chY = coil_assignment['chY']
        chZ = coil_assignment['chZ']

        Ix = Bx / Bslope_TperA[chX]
        Iy = By / Bslope_TperA[chY]
        Iz = Bz / Bslope_TperA[chZ]

        logger.info(f'Setting {Bmag=}, {theta=} {phi=} in radians')
        logger.info(f"Setting {chX=} to {Ix}")
        logger.info(f"Setting {chY=} to {Iy}")
        logger.info(f"Setting {chZ=} to {Iz}")
        self.setChanIout(chX, Ix)
        self.setChanIout(chY, Iy)
        self.setChanIout(chZ, Iz)
        return Ix, Iy, Iz

    def setBinDegrees(self, Bmag=50e-6, theta=0, phi=0):
        logger.info(f'Setting {Bmag=},  {theta=} {phi=} in degrees')
        return self.setB(Bmag=Bmag, theta=theta/180*np.pi, phi=phi/180*np.pi)

# TSDB logger setting
tsdb_client = tsdb.Client('influx', 'http://qo.physics.wm.edu:8428', database='qolab')
tsdb_ingester = tsdb.Ingester(tsdb_client, batch=11, measurement_prefix='VAMPIRE.VCSEL')

logger.info("Accessing hardware")
rm = pyvisa.ResourceManager()
instr=rm.open_resource('USB0::10893::4354::MY61001869::0::INSTR')
app_nickname = 'BfieldDriver'
#Bfield = BfieldDriver(instr, device_nickname='.'.join([app_nickname, 'b_field_driver']), tsdb_ingester=tsdb_ingester)
Bfield = BfieldDriver(instr, device_nickname='.'.join([app_nickname, 'b_field_driver']), tsdb_ingester=None)


print('Use me as : Bfield.setBinDegrees(theta=Angle, phi=Angle)')




from qolab.hardware.scpi import SCPIinstr
from qolab.hardware.basic import BasicInstrument

class RFGenerator(BasicInstrument):
    # Minimal set of methods to be implemented by a RFGenerator
    def __init__(self, *args, **kwds):
        BasicInstrument.__init__(self, *args, **kwds)
        self.config['Device type']='RFGenerator'
        self.config['Device model'] = 'Generic RF generator Without Hardware interface'
        self.config['FnamePrefix'] = 'rfgen'
        self.deviceProperties.update({'FreqFixed'})

class RFGeneratorSCPI(SCPIinstr, RFGenerator):
    """     
    Do not instantiate directly, use
    rm = pyvisa.ResourceManager()
    RFGeneratorSCPI(rm.open_resource('TCPIP::192.168.0.2::INSTR'))
    """
    def __init__(self, resource, *args, **kwds):
        SCPIinstr.__init__(self, resource)
        RFGenerator.__init__(self, *args, **kwds)
        self.config['DeviceId'] = str.strip(self.idn)

from .agilent_e8257d import AgilentE8257D
from .qol_lmx2487  import QOL_LMX2487





"""
Created by Eugeniy E. Mikhailov 2021/11/29
"""

from qolab.hardware.basic import BasicInstrument
from qolab.hardware.scope import ScopeSCPI
from qolab.hardware.scpi import response2numStr
from qolab.data.trace import Trace, TraceXY
import numpy as np
import scipy.signal
from pyvisa.constants import InterfaceType

class SDS1104X(ScopeSCPI):
    """ Siglent SDS1104x scope """
    # SDS1104x has actually 8 divisions but its behave like it has 10,
    # the grabbed trace has more points outside what is visible on the screen
    vertDivOnScreen = 10
    horizDivOnScreen = 14
    def __init__(self, resource, *args, **kwds):
        super().__init__(resource, *args, **kwds)
        self.config['Device model'] = 'SDS1104X'
        self.resource.read_termination='\n'
        self.numberOfChannels = 4
        self.maxRequiredPoints = 1000; # desired number of points per channel, can return twice more

    def mean(self, chNum):
        # get mean on a specific channel calculated by scope
        # PAVA stands for PArameter VAlue
        qstr = f'C{chNum}:PAVA? MEAN'
        rstr = self.query(qstr); 
        # reply is in the form 'C1:PAVA MEAN,3.00E-02V'
        prefix, numberString, unit = response2numStr(rstr, firstSeparator=',', unit='V')
        return(float(numberString))

    def getAvailableNumberOfPoints(self, chNum):
        if chNum != 1 and chNum != 3:
            # for whatever reason 'SAMPLE_NUM' fails for channel 2 and 4
            chNum = 1
        qstr = f'SAMPLE_NUM?  C{chNum}'
        rstr = self.query(qstr)
        # reply is in the form 'SANU 7.00E+01pts'
        prefix, numberString, unit = response2numStr(rstr, firstSeparator=' ', unit='pts')
        return(int(float(numberString)))
            
    @BasicInstrument.tsdb_append
    def getSampleRate(self):
        rstr = self.query('SAMPLE_RATE?');
        # expected reply is like 'SARA 1.00E+09Sa/s'
        prefix, numberString, unit = response2numStr(rstr, firstSeparator=' ', unit='Sa/s')
        return(int(float(numberString)))

    def setSampleRate(self, val):
        print('Cannot set SampleRate directly for SDS1104X')
        # it is not possible to do with this model directly
        pass

    def calcSparsingAndNumPoints(self, availableNpnts=None, maxRequiredPoints=None):
        if availableNpnts is None:
            # using channel 1 to get availableNpnts
            availableNpnts = self.getAvailableNumberOfPoints(1)
        if maxRequiredPoints is None:
            maxRequiredPoints = self.maxRequiredPoints

        if availableNpnts <= maxRequiredPoints*2:
            Npnts = availableNpnts
            sparsing = 1
        else:
            sparsing = int(np.floor(availableNpnts/maxRequiredPoints))
            Npnts = int(np.floor(availableNpnts/sparsing))
        return(sparsing, Npnts, availableNpnts, maxRequiredPoints)

    def getRawWaveform(self, chNum, availableNpnts=None, maxRequiredPoints=None, decimate=True):
        """
        If decimate=True is used, we get all available points and then low-pass filter them.
          The result is less noisy. But transfer time from the instrument is longer.
        If decimate=False, then it we are skipping points to get needed number
          but we might see aliasing, if there is a high frequency noise and sparsing > 1
        """

        (sparsing, Npnts, availableNpnts, maxRequiredPoints) = self.calcSparsingAndNumPoints(availableNpnts, maxRequiredPoints)
        if decimate:
            Npnts = availableNpnts # get all of them and decimate later
        if (sparsing == 1 and Npnts == availableNpnts) or decimate:
            # We are getting all points of the trace
            # Apparently sparsing has no effect with this command
            # and effectively uses SP=1 for any sparsing
            # but I want to make sure and force it
            cstr = f'WAVEFORM_SETUP NP,0,FP,0,SP,1'
            # technically when we know Npnts and sparsing
            # we can use command from the follow up 'else' clause
        else:
            # we just ask every point with 'sparsing' interval
            # fast to grab but we could do better with more advance decimate
            # method, which allow better precision for the price of longer acquisition time
            cstr = f'WAVEFORM_SETUP SP,{sparsing},NP,{Npnts},FP,0'
            # Note: it is not enough to provide sparsing (SP),
            # number of points (NP) needed to be calculated properly too!
            # From the manual
            # WAVEFORM_SETUP SP,<sparsing>,NP,<number>,FP,<point> 
            # SP Sparse point. It defines the interval between data points.
            # For example:
            #   SP = 0 sends all data points.
            #   SP = 1 sends all data points.
            #   SP = 4 sends every 4th data point
            # NP — Number of points. It indicates how many points should be transmitted.
            # For example:
            #   NP = 0 sends all data points.
            #   NP = 50 sends a maximum of 50 data points.
            # FP — First point. It specifies the address of the first data point to be sent.
            # For example:
            #   FP = 0 corresponds to the first data point.
            #   FP = 1 corresponds to the second data point
        self.write(cstr)

        trRaw = Trace(f'Ch{chNum}')

        qstr = f'C{chNum}:WAVEFORM? DAT2'
        # expected full reply: 'C1:WF DAT2,#9000000140.........'
        try:
            wfRaw=self.query_binary_values(qstr, datatype='b', header_fmt='ieee', container=np.array, chunk_size=(Npnts+100))
            if self.resource.interface_type == InterfaceType.usb:
                # Somehow on windows (at least with USB interface)
                # there is a lingering empty string which we need to flush out
                r=self.read()
                if r != '':
                    print(f'WARNING: We expected an empty string but got {r=}')
            trRaw.values = wfRaw.reshape(wfRaw.size,1)
            if decimate and sparsing != 1:
                numtaps = 3; # not sure it is the best case
                trRaw.values = scipy.signal.decimate(trRaw.values, sparsing, numtaps, axis=0)
        except ValueError as err:
            # most likely we get crazy number of points
            # self.read() # flushing the bogus output of previous command
            print(f'Error {err=}: getting waveform failed for {qstr=}')
            wfRaw=np.array([])
        trRaw.config['unit'] = 'Count'
        trRaw.config['tags']['Decimate'] = decimate
        return(trRaw, availableNpnts, Npnts)

    def getChanVoltsPerDiv(self, chNum):
        qstr = f'C{chNum}:VDIV?'
        rstr = self.query(qstr)
        # expected reply to query: 'C1:VDIV 1.04E+00V'
        prefix, numberString, unit = response2numStr(rstr, firstSeparator=' ', unit='V')
        return(float(numberString))

    def setChanVoltsPerDiv(self, chNum, vPerDiv):
        cstr = f'C{chNum}:VDIV {vPerDiv}'
        self.write(cstr)
        # if out of range, the VAB bit (bit 2) in the STB register to be set

    def getChanVoltageOffset(self, chNum):
        qstr = f'C{chNum}:OFST?'
        rstr = self.query(qstr)
        # expected reply to query: 'C1:OFST -1.27E+00V'
        prefix, numberString, unit = response2numStr(rstr, firstSeparator=' ', unit='V')
        return(float(numberString))

    def setChanVoltageOffset(self, chNum, val):
        cstr = f'C{chNum}:OFST {val}'
        self.write(cstr)

    def getLED(self):
        """ Returns binary mask of available LEDs """
        qstr = 'LED?'
        rstr = self.query(qstr)
        prefix, numberString, unit = response2numStr(rstr, firstSeparator=' ', unit='')
        return(int(numberString,16)) # convert from hex string to integer

    def toggleRun(self):
        # SY_FP is undocumented, reverse engineered from the web interface
        self.write('SY_FP 12,1')

    @BasicInstrument.tsdb_append
    def getRun(self):
        ledStatus = self.getLED()
        return bool(ledStatus & (1 << 17))

    @BasicInstrument.tsdb_append
    def setRun(self, val):
        state = self.getRun()
        if state != val:
            self.toggleRun()

    @BasicInstrument.tsdb_append
    def getRoll(self):
        ledStatus = self.getLED()
        return bool(ledStatus & (1 << 10))

    def toggleRoll(self):
        # SY_FP is undocumented, reverse engineered from the web interface
        self.write('SY_FP 49,1')

    @BasicInstrument.tsdb_append
    def setRoll(self, val):
        rollState = self.getRoll()
        if rollState != val:
            self.toggleRoll()

    @BasicInstrument.tsdb_append
    def getTimePerDiv(self):
        qstr = 'TDIV?'
        rstr = self.query(qstr)
        # expected reply to query: 'TDIV 2.00E-08S'
        prefix, numberString, unit = response2numStr(rstr, firstSeparator=' ', unit='S')
        return(float(numberString))

    @BasicInstrument.tsdb_append
    def setTimePerDiv(self, timePerDiv):
        cstr = f'TDIV {timePerDiv}'
        self.write(cstr)
        # if out of range, the VAB bit (bit 2) in the STB register to be set

    @BasicInstrument.tsdb_append
    def getTrigDelay(self):
        qstr = 'TRIG_DELAY?'
        rstr = self.query(qstr)
        # expected reply to query: 'TRDL -0.00E+00S'
        prefix, numberString, unit = response2numStr(rstr, firstSeparator=' ', unit='S')
        return(float(numberString))


    def getWaveform(self, chNum, availableNpnts=None, maxRequiredPoints=None, decimate=True):
        """
        For decimate use see getRawWaveform. In short decimate=True is slower but more precise.
        """
        trRaw, availableNpnts, Npnts = self.getRawWaveform(chNum, availableNpnts=availableNpnts, maxRequiredPoints=maxRequiredPoints, decimate=decimate)
        VoltageOffset = self.getChanVoltageOffset(chNum)
        VoltsPerDiv = self.getChanVoltsPerDiv(chNum)
        tr = trRaw
        tr.values = trRaw.values * VoltsPerDiv * self.vertDivOnScreen/250 -VoltageOffset
        tr.config['unit'] = 'Volt'
        tr.config['tags']['VoltageOffset'] = VoltageOffset
        tr.config['tags']['VoltsPerDiv'] = VoltsPerDiv
        return(tr, availableNpnts)

    def getTimeTrace(self, availableNpnts=None, maxRequiredPoints=None):
        (sparsing, Npnts, availableNpnts, maxRequiredPoints) = self.calcSparsingAndNumPoints(availableNpnts, maxRequiredPoints)
        sampleRate = self.getSampleRate()
        timePerDiv = self.getTimePerDiv()
        trigDelay  = self.getTrigDelay()
        if Npnts is None and sparsing is None:
            # using channel 1 as reference
            Npnts = self.getAvailableNumberOfPoints(1)
        tval = np.arange(Npnts) / sampleRate * sparsing;
        tval = tval - timePerDiv * self.horizDivOnScreen/2 - trigDelay
        t = Trace('time')
        t.values = tval.reshape(tval.size,1)
        t.config['unit'] = 'S'
        t.config['tags']['TimePerDiv'] = timePerDiv
        t.config['tags']['TrigDelay']  = trigDelay
        t.config['tags']['SampleRate'] = sampleRate
        t.config['tags']['AvailableNPnts']   = availableNpnts
        t.config['tags']['Npnts']   = availableNpnts
        t.config['tags']['Sparsing']   = sparsing
        return(t)

    def getTriggerMode(self):
        # we expect NORM, AUTO, SINGLE, STOP
        res = self.query('TRIG_MODE?')
        # res is in the form 'TRMD AUTO'
        return res[5:]

    def setTriggerMode(self, val):
        # we expect NORM, AUTO, SINGLE, STOP
        self.write(f'TRIG_MODE {val}')
        
    def getTrace(self, chNum, availableNpnts=None, maxRequiredPoints=None, decimate=True):
        old_trg_mode = self.getTriggerMode()
        self.setTriggerMode('STOP'); # to get synchronous channels
        wfVoltage, availableNpnts = self.getWaveform( chNum, availableNpnts=availableNpnts, maxRequiredPoints=maxRequiredPoints, decimate=decimate)
        t = self.getTimeTrace(availableNpnts=availableNpnts, maxRequiredPoints=maxRequiredPoints)
        tr = TraceXY( f'Ch{chNum}' )
        tr.x = t
        tr.y = wfVoltage
        # restore scope to the before acquisition mode
        if old_trg_mode != "STOP":
            # short speed up here with this check
            self.setTriggerMode(old_trg_mode)
        return( tr )


if __name__ == '__main__': 
    import pyvisa
    print("testing")
    rm = pyvisa.ResourceManager()
    print(rm.list_resources())    
    instr=rm.open_resource('TCPIP::192.168.0.62::INSTR')
    scope = SDS1104X(instr)
    print(f'ID: {scope.idn}')
    # print(f'Ch1 mean: {scope.mean(1)}')
    print(f'Ch1 available points: {scope.getAvailableNumberOfPoints(1)}')
    print(f'Sample Rate: {scope.getSampleRate()}')
    print(f'Time per Div: {scope.getTimePerDiv()}')
    print(f'Ch1 Volts per Div: {scope.getChanVoltsPerDiv(1)}')
    print(f'Ch1 Voltage Offset: {scope.getChanVoltageOffset(1)}')
    print('------ Header start -------------')
    print(str.join('\n', scope.getHeader()))
    print('------ Header ends  -------------')
    ch1  = scope.getTrace(1)
    traces = scope.getAllTraces()
    





import logging
import universal_tsdb as utsdb
from universal_tsdb import Client, MaxErrorsException
import functools
import time

__all__ = [
    'Client', 'Ingester', 'MaxErrorsException'
]

logging.basicConfig(format='%(asctime)s %(levelname)8s %(name)s: %(message)s', datefmt='%m/%d/%Y %H:%M:%S')
logger = logging.getLogger('qolab.tsdb')
logger.setLevel(logging.INFO)

class Ingester(utsdb.Ingester):
    """ Same as universal_tsdb.Ingester but sets measurement_prefix.
        so measurement becomes measurement_prefix.measurement """
    def __init__(self, client, batch=0, measurement_prefix=''):
        super().__init__(client, batch=batch)
        self.measurement_prefix=measurement_prefix

    def append(self, timestamp=None, tags=None, measurement=None, **kwargs):
        if self.measurement_prefix is None or not isinstance(self.measurement_prefix, str):
            raise ValueError('Invalid measurement_prefix, it should be string')
        if measurement is None or not isinstance(measurement, str):
            raise ValueError('Invalid measurement, it should be string')
        qolab_measurement = '.'.join((self.measurement_prefix, measurement))
        # space is illegal for measurements fields
        qolab_measurement=qolab_measurement.replace(' ', '-')
        logger.debug(f'{qolab_measurement=} {tags=}, {kwargs=}')
        return super().append(timestamp=timestamp, tags=tags, measurement=qolab_measurement, **kwargs)

def tsdb_append_metric_for_class_setter_or_getter(tsdb_logger=None):
    def wrap(f):
        @functools.wraps(f)
        def wrapper(*args, **kwds):
            if f.__name__[0:3] != 'get' and  f.__name__[0:3] != 'set':
                logger.warning(f'Do not know how to work with {f.__name__}, it is neither set... or get...')
                ret = f(*args, **kwds)
                return ret

            cls = args[0]
            action = f.__name__[0:3]
            var_name = f.__name__[3:]
            val = None
            if cls.config['DeviceNickname'] is not None:
                device_type=cls.config['DeviceNickname']
            else:
                device_type=cls.config['Device type']
            if action == 'get':
                """ getter """
                val = f(*args, **kwds)
                ts = time.time()
                ret = val
            else:
                """ setter """
                val = args[1]
                ts = time.time()
                ret = f(*args, **kwds)

            logger.debug(f'function {f.__name__} {action} {var_name} = {val}')
            ts_ms = int(ts*1000)
            fields={var_name: val}
            try:
                if cls.tsdb_ingester is not None:
                    cls.tsdb_ingester.append(ts_ms, measurement=device_type, tags={'action': action}, **fields)
            except ValueError as err:
                logger.error(f'{err=} in function {f.__name__}: {var_name} = {val}')
            return ret
        return wrapper
    return wrap

if __name__ == '__main__': 
    from qolab.hardware.basic import BasicInstrument
    tsdb_client = Client('influx', 'http://localhost:8428', database='qolab')
    tsdb_ingester = Ingester(tsdb_client, batch=10, measurement_prefix='experiment.title')
    class InstWithLog(BasicInstrument):
        def __init__(self, *args, **kwds):
            super().__init__(*args, **kwds)
            self.config['Device type'] = 'TestTSDBLogger'
            self.config['Device model'] = 'v01'
            self.config['FnamePrefix'] = 'test_log'
            self.config['SavePath'] = './data'
            self.deviceProperties.update({'D'})
            self.d = 13.45


        @BasicInstrument.tsdb_append
        def setD(self, val):
            self.d=val

        @BasicInstrument.tsdb_append
        def getD(self):
            """ get D variable """
            return self.d

    dev = InstWithLog(tsdb_ingester=tsdb_ingester, device_nickname='tester')
    dev.getD()
    dev.setD(3)

    tsdb_ingester.commit()



import pyvisa
from qolab.hardware.scope import SDS1104X

if __name__ == '__main__': 
    print("Testing SDS1104X")
    rm = pyvisa.ResourceManager()
    print(rm.list_resources())    
    instr=rm.open_resource('TCPIP::192.168.0.61::INSTR')
    scope = SDS1104X(instr)
    print(f'ID: {scope.idn}')
    print(f'Ch1 mean: {scope.mean(1)}')
    print(f'Ch1 available points: {scope.getAvailableNumberOfPoints(1)}')
    print(f'Sample Rate: {scope.getSampleRate()}')
    print(f'Time per Div: {scope.getTimePerDiv()}')
    print(f'Ch1 Volts per Div: {scope.getChanVoltsPerDiv(1)}')
    print(f'Ch1 Voltage Offset: {scope.getChanOffset(1)}')
    ch1  = scope.getTrace(1)
    ch1.plot()




import time
from qolab.hardware.basic import BasicInstrument

class PID(BasicInstrument):
    def __init__(self, Gp=0, Gi=0, Gd=0, sign=1, enable=True, *args, **kwds):
        super().__init__(*args, **kwds)
        self.config['Device model'] = 'Generic Software PID loop'
        self.config['Device type']='PID loop'
        self.config['FnamePrefix'] = 'pid'
        self.deviceProperties.update({'Gp', 'Gi', 'Gd', 'Sign', 'Enable'})
        self.setGp(Gp)
        self.setGi(Gi)
        self.setGd(Gd)
        self.setSign(sign)
        self.setEnable(enable)

        self.reset()

    def __repr__(self):
        s = ''
        s += f'{self.__class__.__name__}('
        s += f'Gp={self.Gp}'
        s += f', Gi={self.Gi}'
        s += f', Gd={self.Gd}'
        s += f', sign={self.enable}'
        if self.config['DeviceNickname'] is not None:
            s += ", device_nickname='"
            s += f"{self.config['DeviceNickname']}"
            s += "'"
        s += ')'
        return(s)

    @BasicInstrument.tsdb_append
    def getGp(self):
        return self.Gp

    @BasicInstrument.tsdb_append
    def setGp(self,val):
        self.Gp=val

    @BasicInstrument.tsdb_append
    def getGi(self):
        return self.Gi

    @BasicInstrument.tsdb_append
    def setGi(self,val):
        self.Gi=val

    @BasicInstrument.tsdb_append
    def getGd(self):
        return self.Gd

    @BasicInstrument.tsdb_append
    def setGd(self,val):
        self.Gd=val

    @BasicInstrument.tsdb_append
    def getSign(self):
        return self.sign

    @BasicInstrument.tsdb_append
    def setSign(self,val):
        self.sign=val

    @BasicInstrument.tsdb_append
    def getEnable(self):
        return self.enable

    @BasicInstrument.tsdb_append
    def setEnable(self,val):
        self.enable=val

    
    def reset(self):
        self.err_1dt_back = 0
        self.err_2dt_back = 0
        self.err_now      = 0
        self.last_update  = time.time()

    def feedback(self, err):
        # PID feedback
        # see https://en.wikipedia.org/wiki/PID_controller#Pseudocode
        self.err_2dt_back = self.err_1dt_back;
        self.err_1dt_back = self.err_now;
        self.err_now = err

        tnow = time.time()
        dt = tnow - self.last_update
        self.last_update = tnow
        A0= self.Gp+self.Gi*dt + self.Gd/dt;
        A1= -self.Gp-2*self.Gd/dt;
        A2= self.Gd/dt;

        u = A0*self.err_now + A1*self.err_1dt_back + A2*self.err_2dt_back;
        u *= self.sign
        if not self.enable:
            return 0
        return u




import yaml
import time
from qolab.file_utils import get_next_data_file
from qolab.tsdb import tsdb_append_metric_for_class_setter_or_getter

class BasicInstrument:
    """ This is the most basic instrument class """
    def __init__(self, config={}, device_nickname=None, tsdb_ingester=None):
        """ 
        device_nickname - used to separate similar instruments,
                          also used as a tag in Time Series DB (TSDB) logging,
                          if not set we fallbacl to 'Device type'
        tsdb_ingester   - used to log to TSDB if not set, the log entry is skipped
        config          - used to add or override default entries
        """
        self.config = {}
        self.config['Device type'] = 'Basic Instrument'
        self.config['Device model'] = 'Model is unset'
        self.config['DeviceId'] = None
        self.config['DeviceNickname'] = device_nickname; # to separate similar instruments
        self.config['FnamePrefix'] = 'basicInstrument'
        self.config['SavePath'] = './data'
        for k, v in config.items():
            self.config[k]=v
        self.tsdb_ingester = tsdb_ingester
        # deviceProperties must have 'get' and preferably 'set' methods available,
        # i.e. 'SampleRate' needs getSampleRate() and love to have setSampleRate(value)
        # they will be used to obtain config and set device according to it
        # self.deviceProperties = {'SampleRate', 'TimePerDiv', 'TrigDelay', };
        self.deviceProperties = {'TimeStamp'}

    def __repr__(self):
        s = ''
        s += f'{self.__class__.__name__}('
        s += 'config={'
        cstr=[]
        for k, v in self.config.items():
            if v is not None:
                cstr.append(
                    f"'{k}'"
                    ': '
                    f"'{v}'")
        s += ', '.join(cstr)
        s += '}'
        if self.tsdb_ingester is not None:
            s += f', tsdb_ingester={self.tsdb_ingester}'
        s += ')'
        return(s)

    def tsdb_append(f):
        return tsdb_append_metric_for_class_setter_or_getter()(f)

    def getChannelsConfig(self):
        chconfig = {}
        if not hasattr(self, 'numberOfChannels'):
            return chconfig

        for chNum in range(1, self.numberOfChannels+1):
            chNconfig = {}
            for p in self.channelProperties:
                getter = f'getChan{p}'
                if  not hasattr(self, getter):
                    print(f'warning no getter for {p}, i.e. {getter} is missing')
                    continue
                res = getattr(self, getter)(chNum) 
                chNconfig[p] = res
            chconfig[chNum] = chNconfig
        return chconfig

    def getConfig(self):
        config = self.config.copy()
        dconfig = {}
        for p in self.deviceProperties:
            if hasattr(self, p) or hasattr(type(self), p):
                dconfig[p] = getattr(self, p)
                continue
            getter = f'get{p}'
            if  not hasattr(self, getter):
                print(f'warning no getter for {p}, i.e. {getter} is missing')
                continue
            res = getattr(self, getter)() 
            dconfig[p] = res
        config['DeviceConfig'] = dconfig
        if not hasattr(self, 'numberOfChannels'):
            return config
        config['ChannelsConfig']=self.getChannelsConfig();
        return config

    def setConfig(self, cfg):
        new_config = cfg.copy()
        device_config = new_config.pop('DeviceConfig')
        self.config.update(new_config)
        for p, v in device_config.items():
            setter = f'set{p}'
            if  hasattr(self, setter):
                # we have something like setProperty
                getattr(self, setter)(v) 
                continue
            if hasattr(self, p) or hasattr(type(self), p):
                # we have attribute Property
                setattr(self, p, v)
                continue
            print(f'warning: both {setter} and attribute {p} are missing, skipping {p}')
            # self.deviceProperties.add(p)
            # setattr(self, p, v)

    def getTimeStamp(self):
        return  time.strftime("%Y/%m/%d %H:%M:%S")

    def getHeader(self):
        header = yaml.dump(self.getConfig(), default_flow_style=False, sort_keys=False)
        header = header.split('\n')
        return header
    
    def getNextDataFile(self, extension='dat'):
        fname = get_next_data_file(self.config['FnamePrefix'], self.config['SavePath'], extension=extension)
        return fname




"""
provide basic class to operate SCPI capable instruments
"""

import re
import logging
import time

logging.basicConfig(format='%(asctime)s %(levelname)8s %(name)s: %(message)s', datefmt='%m/%d/%Y %H:%M:%S')
logger = logging.getLogger('qolab.hardware.scpi')
logger.setLevel(logging.INFO)

def response2numStr(strIn, firstSeparator=None, unit=None):
    # Often an instrument reply is in the form 'TDIV 2.00E-08S' (for example Siglent Scope)
    # i.e. "<prefix><firstSeparator><numberString><unit>
    # prefix='TDIV', firstSeparator=' ', numberString='2.00E-08', unit='S'
    # this function parses the reply
    prefix = None
    rstr = strIn
    if firstSeparator is not None and firstSeparator != '':
        spltStr = re.split(firstSeparator, strIn)
        prefix = spltStr[0]
        rstr = spltStr[1]
    if unit is not None and unit != '':
        spltStr = re.split(unit, rstr)
        if len(spltStr) == 1:
            unit = None
        numberString = spltStr[0]
    else:
        numberString = rstr
    return (prefix, numberString, unit)

class SCPI_PROPERTY(property):
    """ 
    Overrides property class and makes it suitable for SCPI set and query notation.
    Works within SCPIinstr class since it assumes that owner has query() and write().
        scpi_prfx - SCPI command prefix to get/set property, for example 'FreqInt'
                    is internally transformed to 
                    query 'FreqInt?' and setter 'FreqInt {val}'.
                    It could be set as the explicit query and set format string list:
                    ['AUXV? 1', 'AUXV 1,{}'] where {} is place holder for set value
        ptype     - property type 'str', 'int', 'float', ...
        doc       - short description of property, for example 'Internal lockin frequency'
        no_getter - typical for some sort of command
        no_setter - typical for measured values (i.e. external to the instrument)
        tsdb_logging  - logging to time series DB (True/False)
    """
    def __init__(self, scpi_prfx=None, ptype=str, doc=None, no_getter=False, no_setter=False, tsdb_logging=True):
        self.no_getter = no_getter
        self.no_setter = no_setter
        self.tsdb_logging = tsdb_logging
        if no_getter:
            fget=None
        else:
            fget = self.get_scpi
        if no_setter:
            fset=None
        else:
            fset = self.set_scpi
        super().__init__(fget=fget, fset=fset)
        self.scpi_prfx = scpi_prfx
        self.ptype = ptype
        self.__doc__ = doc
        if isinstance(scpi_prfx, str):
            self.scpi_prfx_get = ''.join([self.scpi_prfx, '?'])
            self.scpi_prfx_set = ''.join([self.scpi_prfx, ' {}'])
        elif isinstance(scpi_prfx, list):
            if len(scpi_prfx) != 2:
                raise ValueError(f'{scpi_prfx=}, should be list with exactly two elements')
            self.scpi_prfx_get = self.scpi_prfx[0]
            self.scpi_prfx_set = self.scpi_prfx[1]
        else:
            raise ValueError(f'{scpi_prfx=}, it should be either str or list type')

        if not isinstance(self.scpi_prfx_get, str):
            raise ValueError(f'{self.scpi_prfx_get=}, it should be str type')
        if not isinstance(self.scpi_prfx_set, str):
            raise ValueError(f'{self.scpi_prfx_set=}, it should be str type')

    def __set_name__(self, owner, name):
        self.public_name = name
        self.private_name = '_' + name

    def log_to_tsdb(self, owner, action=None, val=None):
        if owner.tsdb_ingester is None or not self.tsdb_logging:
            return
        if owner.config['DeviceNickname'] is not None:
            measurement=owner.config['DeviceNickname']
        else:
            measurement=owner.config['Device type']
        ts = time.time()
        ts_ms = int(ts*1000)
        var_name = self.public_name
        tags = {'action': action }
        fields = {var_name: val}
        try:
            msg=f"{ts_ms=}, {measurement=}, {tags=}, {fields=}"
            logger.debug(msg)
            owner.tsdb_ingester.append(ts_ms, measurement=measurement, tags=tags, **fields)
        except ValueError as err:
            logger.error(f'{err=}: cannot log to TSDB {var_name} = {val}')


    def get_scpi(self, owner):
        val = self.ptype( owner.query(f'{self.scpi_prfx_get}') )
        self.log_to_tsdb(owner, action='get', val=val)
        return val

    def set_scpi(self, owner, val):
        cstr = self.scpi_prfx_set.format(val)
        owner.write(cstr)
        self.log_to_tsdb(owner, action='set', val=val)

    def __repr__(self):
        sargs= []
        sargs.append( f'scpi_prfx={self.scpi_prfx}')
        sargs.append( f'ptype={self.ptype}')
        sargs.append( f'doc={self.__doc__}')
        sargs.append( f'no_getter={self.no_getter}')
        sargs.append( f'no_setter={self.no_setter}')
        sargs.append( f'tsdb_logging={self.tsdb_logging}')
        sargs =', '.join(sargs)
        s = ''.join( [ f'{self.__class__.__name__}(' , sargs, ')' ] )
        return s

class SCPIinstr:
    """ Basic class which support SCPI commands """
    """     
    Do not instantiate directly, use
    rm = pyvisa.ResourceManager()
    SCPIinstr(rm.open_resource('TCPIP::192.168.0.2::INSTR'))
    """
    def __init__(self, resource):
        self.resource = resource
        
        # convenience pyvisa functions
        self.write = self.resource.write
        self.read = self.resource.read
        self.query = self.resource.query
        self.read_bytes = self.resource.read_bytes
        self.read_binary_values = self.resource.read_binary_values
        self.query_binary_values = self.resource.query_binary_values

    @property
    def idn(self):
        return self.query("*IDN?")

    def clear_status(self):
        self.write("*CLS")

    def set_event_status_enable(self):
        self.write("*ESE")
    
    def query_event_status_enable(self):
        return self.query("*ESE?")

    def query_event_status_register(self):
        return self.query("*ESR?")

    def set_wait_until_finished(self):
        self.write("*OPC")

    def wait_until_finished(self):
        return self.query("*OPC?")

    def reset(self):
        self.write("*RST")
    
    def set_service_request_enable(self):
        self.write("*SRE")

    def query_service_request_enable(self):
        return self.query("*SRE?")

    def query_status_byte(self):
        return self.query("*STB?")
    
    def self_test_result(self):
        return self.query("*TST?")
    
    def wait(self):
        self.write("*WAI")


if __name__ == '__main__': 
    from qolab.hardware.basic import BasicInstrument
    class DummyInstrument(BasicInstrument):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        # in order to check SCPI_PROPERTY we need to implement write and query
        def write(self, str):
            print(f'write: {str=}')

        def query(self, str):
            print(f'query: {str=}')
            return '123'

        x = SCPI_PROPERTY(scpi_prfx='SETX', ptype=str, doc='property X', tsdb_logging=False)
        y = SCPI_PROPERTY(scpi_prfx='SETY', ptype=int, no_setter=True, doc='property Y')
        z = SCPI_PROPERTY(scpi_prfx='SETY', ptype=int, no_getter=True, doc='property Z')

    c1= DummyInstrument()
    c1.deviceProperties.update({'x', 'y'})
    c1.getConfig()
    c2= DummyInstrument()
        



from qolab.hardware.basic import BasicInstrument
from qolab.hardware.lockin import LockinSCPI
from qolab.hardware.scpi  import SCPI_PROPERTY

class SRS_SR865A(LockinSCPI):
    """ SRS SR865A lockin """
    def __init__(self, resource, *args, **kwds):
        super().__init__(resource, *args, **kwds)
        self.config['Device model']='SRS SR865A'
        self.resource.read_termination='\n'
        self.deviceProperties.update({'TimeBaseMode', 'AuxOut1', 'AuxOut2', 'AuxOut3', 'AuxOut4'})

    FreqInt = SCPI_PROPERTY(scpi_prfx='FreqInt', ptype=float, doc='Internal LO frequency')
    FreqExt = SCPI_PROPERTY(scpi_prfx='FreqExt', ptype=float, doc='External LO frequency', no_setter=True)
    Harm = SCPI_PROPERTY(scpi_prfx='Harm', ptype=float, doc='Harmonic of the LO')
    SinAmpl = SCPI_PROPERTY(scpi_prfx='SLVL', ptype=float, doc='Modulation amplitude')
    SinOffset = SCPI_PROPERTY(scpi_prfx='SOFF', ptype=float, doc='Modulation offset')
    EquivalentNoiseBW = SCPI_PROPERTY(scpi_prfx='ENBW', ptype=float, doc='Equivalent Noise BandWidth (it is not 3dB BW, see the manual)', no_setter=True)
    TimeBaseMode = SCPI_PROPERTY(scpi_prfx='TBMODE', ptype=int, doc='10 MHz reference mode: 0 - Auto, 1 - Internal')
    RefPhase = SCPI_PROPERTY(scpi_prfx='PHAS', ptype=float, doc='LO/reference phase')

    Sensitivity = SCPI_PROPERTY(scpi_prfx='SCAL', ptype=int, doc="""
    Sensitivity:
        0:   1  V (uA),  6:  10 mV (nA),  ... 27: 1 nV (fA)
        1: 500 mV (nA),  7:   5 mV (nA),
        2: 200 mV (nA),  8:   2 mV (nA),
        3: 100 mV (nA),  9:   1 mV (nA),
        4:  50 mV (nA), 10: 500 uV (pA),
        5:  20 mV (nA), 11: 200 uV (pA),
    """)
    TimeConstan = SCPI_PROPERTY(scpi_prfx='OFLT', ptype=int, doc="""
    Time constant:
        0:   1 uS,    6:   1 mS, ..... , 21: 30 kS
        1:   3 uS,    7:   3 mS,
        2:  10 uS,    8:  10 mS,
        3:  30 uS,    9:  30 mS,
        4: 100 uS,   10: 100 mS,
        5: 300 uS,   11: 300 mS, 
    """)
    FilterSlope = SCPI_PROPERTY(scpi_prfx='OFSL', ptype=int, doc="""
    Output Filter slope:
        0:  6dB/Oct
        1: 12dB/Oct
        2: 18dB/Oct
        3: 24dB/Oct
    """)

    AuxOut1 = SCPI_PROPERTY(scpi_prfx=['AUXV? 0', 'AUXV 0,{}'], ptype=float, doc='Voltage at Auxilarly output 1')
    AuxOut2 = SCPI_PROPERTY(scpi_prfx=['AUXV? 1', 'AUXV 1,{}'], ptype=float, doc='Voltage at Auxilarly output 2')
    AuxOut3 = SCPI_PROPERTY(scpi_prfx=['AUXV? 2', 'AUXV 2,{}'], ptype=float, doc='Voltage at Auxilarly output 3')
    AuxOut4 = SCPI_PROPERTY(scpi_prfx=['AUXV? 3', 'AUXV 3,{}'], ptype=float, doc='Voltage at Auxilarly output 4')

    AuxIn1 = SCPI_PROPERTY(scpi_prfx=['OAUX? 0', ''], ptype=float, doc='Voltage at Auxilarly input 1', no_setter=True)
    AuxIn2 = SCPI_PROPERTY(scpi_prfx=['OAUX? 1', ''], ptype=float, doc='Voltage at Auxilarly input 2', no_setter=True)
    AuxIn3 = SCPI_PROPERTY(scpi_prfx=['OAUX? 2', ''], ptype=float, doc='Voltage at Auxilarly input 3', no_setter=True)
    AuxIn4 = SCPI_PROPERTY(scpi_prfx=['OAUX? 3', ''], ptype=float, doc='Voltage at Auxilarly input 4', no_setter=True)

if __name__ == '__main__': 
    import pyvisa
    print("testing")
    rm = pyvisa.ResourceManager()
    print(rm.list_resources())    
    instr=rm.open_resource('TCPIP::192.168.0.51::INSTR')
    lockin = SRS_SR865A(instr)
    print('------ Header start -------------')
    print(str.join('\n', lockin.getHeader()))
    print('------ Header ends  -------------')





"""
by Rob Behary and Eugeniy Mikhailov 2024/06/17
"""

from qolab.hardware.scope import SDS1104X
from qolab.hardware.basic import BasicInstrument

import re

class SDS2304X(SDS1104X):
    """ Siglent SDS2304x scope """
    vertDivOnScreen = 8 # unlike SDS1104X, number of divisions matches what is seen on the screen
    horizDivOnScreen = 10
    def __init__(self, resource, *args, **kwds):
        super().__init__(resource, *args, **kwds)
        self.config['Device model'] = 'SDS2304X'
        self.resource.read_termination='\n'
        self.numberOfChannels = 4
        self.maxRequiredPoints = 1000; # desired number of points per channel, can return twice more


    @BasicInstrument.tsdb_append
    def getTimePerDiv(self):
        qstr = 'TDIV?'
        rstr = self.query(qstr)
        # Siglent claims that this model should have same commands as SDS1104X
        # however response is different, for example we got '2.00E-08S' instead 'TDIV 2.00E-08S'
        # expected reply to query: '2.00E-08S' 
        prefix, numberString, unit = response2numStr(rstr, firstSeparator=None, unit='S')
        return(float(numberString))




from qolab.data.trace import loadTrace
import numpy as np

def test_load_uncompressed_v0dot1_trace():
    tr = loadTrace('tests/trace_test_data/xtrace1.dat')
    cfg = tr.getConfig()
    assert cfg['config']['version'] == '0.1'
    assert cfg['config']['model'] == 'Trace'
    data = tr.getData()
    assert np.all( (data - np.array([[1], [3], [2], [5]])) == 0 )

def test_load_gzip_compressed_v0dot1_trace():
    tr = loadTrace('tests/trace_test_data/xtrace1.dat.gz')
    cfg = tr.getConfig()
    assert cfg['config']['version'] == '0.1'
    assert cfg['config']['model'] == 'Trace'
    data = tr.getData()
    assert np.all( (data - np.array([[1], [3], [2], [5]])) == 0 )




from qolab.file_utils import save_table_with_header
import datetime
import numpy as np
import yaml
import pandas

def headerFromDictionary(d, prefix=''):
    header = []
    tail = yaml.dump(d, default_flow_style=False, sort_keys=False)
    tail = tail.split('\n')
    header.extend(tail)
    prefixed_header = [prefix+l for l in header]
    return prefixed_header

def from_timestamps_to_dates(timestamps):
    dates = [datetime.datetime.fromtimestamp(float(ts)) for ts in timestamps]
    return(dates)

def loadTraceRawHeaderAndData(fname):
    headerstr=[]
    data = None
    # we will try to guess if the file compressed
    _open = open
    if fname[-3:] == '.gz':
        # TODO improve detection: gzip files have first 2 bytes set to b'\x1f\x8b'
        import gzip
        _open = gzip.open
    with _open(fname, mode='rb') as tracefile:
        # Reading yaml header prefixed by '% '
        # It sits at the top and below is just data in TSV format
        while True:
            ln = tracefile.readline()
            if ln[0:2]==b'% ':
                headerstr.append(ln[2:].decode('utf-8'))
            else:
                break
        header=yaml.load(str.join('\n', headerstr), Loader=yaml.BaseLoader)
        # now we load the data itself
        # data=np.genfromtxt(fname, comments='%', delimiter='\t')
        # Note: pandas reads csv faster by factor of 8 then numpy.genfromtxt
        # data=pandas.read_csv('/home/evmik/hopping_trace_20220706_02141.dat', comment='%', delimiter='\t', header=None)
        tracefile.seek(0) # rewind file to the beginning
        df = pandas.read_csv(tracefile, comment='%', delimiter='\t', header=None)
        data = df.to_numpy()
    return(header, data)

def loadTrace(fname):
    (header, data) = loadTraceRawHeaderAndData(fname)
    return traceFromHeaderAndData(header, data)

def traceFromHeaderAndData(header, data=None):
    label = None
    model = None
    tr = None
    if 'config' not in header:
        print('Error: trace has now config')
        return None
    else:
        if 'label' in header['config']:
            label = header['config']['label']
        if 'model' not in header['config']:
            print('Error: unknown trace model')
            return None
        else:
            model = header['config']['model']
            if model == 'Trace':
                tr = Trace(label)
                if data is not None:
                    tr.values = data
            elif model == 'TraceXY':
                tx = traceFromHeaderAndData(header['TraceX'])
                ty = traceFromHeaderAndData(header['TraceY'])
                if data is not None:
                    tx.values=data[:,0]
                    ty.values=data[:,1]
                tr = TraceXY(label)
                tr.x = tx
                tr.y = ty
            elif model == 'TraceSetSameX':
                tx = traceFromHeaderAndData(header['TraceX'])
                tx.values=data[:,0]
                tr = TraceSetSameX(label)
                tr.addTraceX(tx)
                ytrs_header = header['TraceY']
                cnt=0
                for l, h in ytrs_header.items():
                    ty = traceFromHeaderAndData(h)
                    cnt += 1
                    ty.values=data[:,cnt]
                    trxy = TraceXY(l)
                    trxy.x = tx
                    trxy.y = ty
                    tr.addTrace( trxy )

            else:
                print(f'Error: unknown trace model: {model}')
                return None
        tr.config = header['config']

    return tr
    
class Trace:
    def __init__(self, label):
        self.config = {}
        self.config['label'] = label
        self.config['model'] = 'Trace'
        self.config['version'] = '0.1'
        # 'type' is useful to indicate way of representation, make sense on y_vs_x traces
        # if set to none we have normal y vs x
        # if set to 'timestamp' x will be converted to datetime dates
        self.config['type'] = None 
        self.config['item_format']='.15e'
        self.config['tags'] = {}
        self.last_saved_pos = 0
        self._trace_specific_init()
        self.clear_data()

    def _trace_specific_init(self):
        self.config['unit'] = None
        self.values = np.empty(0)
        self.last_saved_pos = 0

    def clear_last_saved_pos(self):
        self.last_saved_pos = 0

    def clear_data(self):
        self.clear_last_saved_pos()
        if self.values is not None:
            self.values = np.empty(0, dtype=self.values.dtype)

    def __repr__(self):
        lbl = self.config['label']
        cls_name = f"{self.__class__.__name__}('{lbl}'"
        return "".join([cls_name, f', N={self.values.size}', ')'])

    def plot(self):
        import matplotlib.pyplot as plt
        x=self.values
        if self.config['type'] is not None:
            if self.config['type'] == 'timestamp':
                x = from_timestamps_to_dates(self.values)
        plt.plot(x, label=self.config['label'])
        plt.xlabel('index')
        plt.ylabel(f"{self.config['unit']}")
        plt.legend()
        plt.grid(True)
    
    def getConfig(self):
        d ={}
        d['config'] = self.config.copy()
        return( d )

    def getData(self):
        return( self.values )

    def getHeader(self, prefix=''):
        d = self.getConfig()
        return headerFromDictionary(d, prefix='')

    def save(self, fname, last_saved_pos=None, skip_headers_if_file_exist=False,  **kwargs):
        if last_saved_pos is None:
            last_saved_pos = self.last_saved_pos
        data = self.getData()
        if last_saved_pos > 0:
            skip_headers_if_file_exist=True
        fname = save_table_with_header(fname, data[last_saved_pos:,:], self.getHeader(), item_format=self.config['item_format'], skip_headers_if_file_exist=skip_headers_if_file_exist, **kwargs)
        self.last_saved_pos = data.shape[0]
        return(fname)

    def addPoint(self, val):
        self.values = np.append(self.values, val)


class TraceXY(Trace):
    def __init__(self, label):
        super().__init__(label)
        self.config['model'] = 'TraceXY'

    def _trace_specific_init(self):
        self.x = None
        self.y = None

    def clear_data(self):
        self.clear_last_saved_pos()
        if self.x is not None:
            self.x.clear_data()
        if self.y is not None:
            self.y.clear_data()

    def __repr__(self):
        lbl = self.config['label']
        cls_name = f"{self.__class__.__name__}('{lbl}'"
        xlabel= f"{self.x.config['label']}"
        xparam= f", {self.x}"
        yparam= f", {self.y}"
        return "".join([cls_name, xparam, yparam, ')'])

    def plot(self):
        import matplotlib.pyplot as plt
        x=self.x.values
        if self.x.config['type'] is not None:
            if self.x.config['type'] == 'timestamp':
                x = from_timestamps_to_dates(x)
        plt.plot(x, self.y.values, label=self.config['label'])
        plt.xlabel(f"{self.x.config['label']} ({self.x.config['unit']})")
        plt.ylabel(f"{self.y.config['label']} ({self.y.config['unit']})")
        plt.legend()
        plt.grid(True)

    def getConfig(self):
        config = {}
        config['config'] = self.config.copy()
        config['TraceX'] = {}
        config['TraceX'] = self.x.getConfig()
        config['TraceY'] = {}
        config['TraceY'] = self.y.getConfig()
        return( config )

    def getData(self):
        data=self.x.values
        if data.ndim == 1:
            data = data[:, np.newaxis]
        vals = self.y.values
        if vals.ndim == 1:
            vals = vals[:, np.newaxis]
        data=np.concatenate((data, vals), 1)
        return( data )

    def addPoint(self, valX, valY):
        self.x.values = np.append(self.x.values, valX)
        self.y.values = np.append(self.y.values, valY)

        
class TraceSetSameX(Trace):
    def __init__(self, label):
        super().__init__(label)
        self.config['model'] = 'TraceSetSameX'

    def _trace_specific_init(self):
        self.x = None
        self.traces={}

    def clear_data(self):
        self.clear_last_saved_pos()
        if self.x is not None:
            self.x.clear_data()
        for k, tr in self.traces.items():
            tr.clear_data()

    def __repr__(self):
        lbl = self.config['label']
        cls_name = f"{self.__class__.__name__}('{lbl}'"
        xparam= f", x: {self.x}"
        yparam = f", traces: {list(self.traces.keys())}"
        return "".join([cls_name, xparam, yparam, ')'])

    def addTraceX(self, tr):
        self.x = tr

    def addTrace(self, tr):
        if tr.config['model'] == 'TraceXY':
            if self.x == None:
                self.x = tr.x
            trY = tr.y
            self.traces[tr.config['label']]=trY
        elif tr.config['model'] == 'Trace':
            if self.x == None:
                self.x = tr
            else:
                self.traces[tr.config['label']]=tr
        
    def plot(self):
        import matplotlib.pyplot as plt
        nplots = len(self.traces.keys())
        fig = plt.gcf()
        fig, axs = plt.subplots(nplots, 1, sharex=True, num=fig.number)
        cnt=0
        x=self.x.values
        if self.x.config['type'] is not None:
            if self.x.config['type'] == 'timestamp':
                x = from_timestamps_to_dates(x)
        for k, tr in self.traces.items():
            p=axs[cnt].plot(x, tr.values, label=k)
            axs[cnt].set_ylabel(f"{tr.config['label']} ({tr.config['unit']})")
            axs[cnt].legend()
            axs[cnt].grid(True)
            cnt+=1
        plt.xlabel(f"{self.x.config['label']} ({self.x.config['unit']})")

    def items(self):
        return (self.traces.items())

    def keys(self):
        return (self.traces.keys())

    def getTrace(self, label):
        tr = TraceXY(label)
        tr.x = self.x
        tr.y = self.traces[label]
        return (tr)
    
    def getConfig(self):
        config = {}
        config['config'] = self.config.copy()
        config['TraceX'] = {}
        config['TraceX'] = self.x.getConfig()
        config['TraceY'] = {}
        for k,v in self.traces.items():
            config['TraceY'][k] = v.getConfig()
        return( config )

    def getData(self):
        data=self.x.values
        if data.ndim == 1:
            data = data[:, np.newaxis]
        for k,v in self.traces.items():
            vals = v.values
            if vals.ndim == 1:
                vals = vals[:, np.newaxis]
            data=np.concatenate((data, vals), 1)
        return( data )

    def addPointToTrace(self, val, name=None):
        if name is None:
            self.x.values = np.append(self.x.values, val)
        else:
            a = self.traces[name].values
            a = np.append(a, val)
            self.traces[name].values = a



if __name__ == '__main__': 
    print("Testing trace")
    x=Trace('x trace')
    x.values = np.random.normal(2,2,(4,1))
    x.values = np.array(x.values, int)
    x.config['unit']='s'
    x.config['tags']['tag1'] = 'xxxx'
    x.config['tags']['tag2'] = 'xxxx'
    x.save('xtrace.dat', skip_headers_if_file_exist=True)
    # print(x.getHeader())
    y=Trace('y trace')
    y.values = np.random.normal(2,2,(4,1))
    y.config['unit']='V'
    y.config['tags']['ytag2'] = 'yyyy'
    xy=TraceXY('xy trace')
    xy.config['tags']['xy tag']= 'I am xy tag'
    xy.x = x
    xy.y = y
    xy.save('xytrace.dat')
    # print(xy.getHeader())
    xyn = TraceSetSameX('many ys trace')
    xyn.config['tags']['descr'] = 'I am many ys trace'
    xy.config['label']='y1'
    xyn.addTrace(xy)
    xy.config['label']='y2'
    xyn.addTrace(xy)
    xy.config['label']='y3'
    xyn.addTrace(xy)
    xyn.save('xyntrace.dat')
    # print(xyn.getHeader())





"""
Provide basic class to operate scope
Created by Eugeniy E. Mikhailov 2021/11/29
"""
from qolab.hardware.scpi import SCPIinstr
from qolab.hardware.basic import BasicInstrument
from qolab.data.trace import TraceSetSameX
import yaml

class Scope(BasicInstrument):

    # Minimal set of methods to be implemented by a scope.
    def __init__(self, *args, **kwds):
        BasicInstrument.__init__(self, *args, **kwds)
        self.config['Device type']='Scope'
        self.config['Device model'] = 'Generic Scope Without Hardware interface'
        self.config['FnamePrefix'] = 'scope'
        self.numberOfChannels = 0
        # deviceProperties must have 'get' and preferably 'set' methods available,
        # i.e. 'SampleRate' needs getSampleRate() and love to have setSampleRate(value)
        # they will be used to obtain config and set device according to it
        self.deviceProperties.update({'SampleRate', 'TimePerDiv', 'TrigDelay', 'TriggerMode', 'Roll', 'Run' })
        # same is applied to channelProperties but we need setter/getter with channel number
        # i.e.  VoltsPerDiv ->  getChanVoltsPerDiv(chNum) and setSampleRate(chNum, value)
        self.channelProperties = {'VoltsPerDiv', 'VoltageOffset', }

    def getTrace(self, chNum, availableNpnts=None, maxRequiredPoints=None, decimate=True):
        # Should work with minimal arguments list 
        # but might be faster if parameters provided: less IO requests
        # old_trg_mode = self.getTriggerMode()
        # self.setTriggerMode('STOP'); # to get synchronous channels
        raise NotImplementedError( 'getTrace function is not implemented' )
        # if old_trg_mode != "STOP":
            # short speed up here with this check
            # self.setTriggerMode(old_trg_mode)

    def getTriggerMode(self):
        # we expect NORM, AUTO, SINGLE, STOP
        raise NotImplementedError( 'getTriggerMode function is not implemented' )

    def setTriggerMode(self, mode):
        # we expect NORM, AUTO, SINGLE, STOP
        raise NotImplementedError( 'setTriggerMode function is not implemented' )

    def getAllTraces(self, availableNpnts=None, maxRequiredPoints=None, decimate=True):
        allTraces=TraceSetSameX('scope traces')
        allTraces.config['tags']['DAQ']=self.getConfig()
        old_trg_mode = self.getTriggerMode()
        self.setTriggerMode('STOP'); # to get synchronous channels
        for chNum in range(1, self.numberOfChannels+1):
            allTraces.addTrace( self.getTrace(chNum, availableNpnts=availableNpnts, maxRequiredPoints=maxRequiredPoints, decimate=decimate) )
        # restore scope to the before acquisition mode
        if old_trg_mode != "STOP":
            # short speed up here with this check
            self.setTriggerMode(old_trg_mode)
        return( allTraces )

    def plot(self, **kwargs):
        allTraces=self.getAllTraces(**kwargs)
        allTraces.plot()

    def save(self, fname=None, item_format='e', availableNpnts=None, maxRequiredPoints=None, decimate=True, extension='dat'):
        allTraces = self.getAllTraces(availableNpnts=availableNpnts, maxRequiredPoints=maxRequiredPoints, decimate=decimate)
        allTraces.config['item_format']=item_format
        if fname is None:
            fname = self.getNextDataFile(extension=extension)
        allTraces.save(fname)
        print(f'Data saved to: {fname}')
        return(fname)


class ScopeSCPI(SCPIinstr, Scope):
    """     
    Do not instantiate directly, use
    rm = pyvisa.ResourceManager()
    ScopeSCPI(rm.open_resource('TCPIP::192.168.0.2::INSTR'))
    """
    def __init__(self, resource, *args, **kwds):
        SCPIinstr.__init__(self, resource)
        Scope.__init__(self, *args, **kwds)
        self.config['DeviceId'] = str.strip(self.idn)

from .sds1104x import SDS1104X
from .sds2304x import SDS2304X




import logging
# this should be done before justpy is called or log formatter does not work
logging.basicConfig(format='%(asctime)s %(levelname)8s %(name)s: %(message)s', datefmt='%m/%d/%Y %H:%M:%S')

import pyvisa
import numpy as np
import platform
import time
from tqdm import tqdm as pbar
import matplotlib.pyplot as plt


from qolab.hardware import BasicInstrument
from qolab.data import TraceSetSameX, TraceXY, Trace
from qolab.hardware.rf_generator import QOL_LMX2487
from qolab.hardware.daq import LabJackUE9
from qolab.hardware.power_supply.keysight_e3612a import KeysightE3612A
import qolab.tsdb as tsdb

l = logging.getLogger('qolab.tsdb')
l.setLevel(logging.INFO)
logger = logging.getLogger('Magnetometer')
logger.setLevel(logging.INFO)

def getConfig(apparatus):
    config = apparatus.config.copy()
    ai = apparatus.instruments
    for n, i in ai.items():
        config[n]=i.getConfig()
    return config

class Apparatus(BasicInstrument):
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)

    def getConfig(self):
        config = self.config.copy()
        ai = self.instruments
        for n, i in ai.items():
            config[n]=i.getConfig()
        return config

class BfieldDriver(KeysightE3612A):
    """ need to set power supply """
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        self.config['Device type'] = 'B field coil driver based on Keysight E3612A power supply'
        self.config['Device model'] = 'v0.1'
        self.deviceProperties.update({'B', 'Bslope_TperA', 'CoilAssignment'})
        """"
        Rough magnetic field calibration of the 3 axes coils in large magnetic shield
        - Ch1: 70mA -> 650 kHz shift for delta m = 2
        - Ch2: 70mA -> 700 kHz shift for delta m = 2
        - Ch2: 70mA -> 659 kHz shift for delta m = 2
        A better calibration obtained on 20220601 see file 
          20220601.magnetic_field_callibration/calibration_currentToB.dat
        """
        # B response to current in a given channel
        self._Bslope_TperA = {
                1: 0.0006574710928926532,
                2: 0.0007064314754023079,
                3: 0.0006635058865577695
                }
        # assuming that Ch1 controls Bz, Ch2 -> By, Ch3 -> Bx
        self._coil_assignment = { 'chX': 3, 'chY': 2, 'chZ': 1 }

    def getBslope_TperA(self):
        return self._Bslope_TperA

    def getCoilAssignment(self):
        return self._coil_assignment

    def getB(self):
        Bslope_TperA = self.getBslope_TperA()
        coil_assignment = self.getCoilAssignment()
        chX = coil_assignment['chX']
        chY = coil_assignment['chY']
        chZ = coil_assignment['chZ']

        Ix = self.getChanIout(chX)
        Iy = self.getChanIout(chY)
        Iz = self.getChanIout(chZ)
        Bx = Ix * Bslope_TperA[chX]
        By = Iy * Bslope_TperA[chY]
        Bz = Iz * Bslope_TperA[chZ]
        Bmag = float(np.sqrt(Bx*Bx + By*By + Bz*Bz))
        theta = float(np.arccos(Bz/Bmag))
        phi  =  float(np.arctan2(By, Bx))
        theta_degree = float(theta/np.pi*180)
        phi_degree = float(phi/np.pi*180)
        return { 'Bmag': Bmag,
                'theta': theta, 'phi': phi,
                'theta_degree': theta_degree, 'phi_degree': phi_degree,
                'Bx': float(Bx),
                'By': float(By),
                'Bz': float(Bz),
                }

    def setB(self, Bmag=50e-6, theta=0, phi=0):
        """ Sets B field currents based on B (in T) and angles theta, and phi """
        self._Bmag = Bmag
        self._theta = theta
        self._phi = phi

        Bx = Bmag*np.sin(theta)*np.cos(phi)
        By = Bmag*np.sin(theta)*np.sin(phi)
        Bz = Bmag*np.cos(theta)

        Bslope_TperA = self.getBslope_TperA()
        coil_assignment = self.getCoilAssignment()
        chX = coil_assignment['chX']
        chY = coil_assignment['chY']
        chZ = coil_assignment['chZ']

        Ix = Bx / Bslope_TperA[chX]
        Iy = By / Bslope_TperA[chY]
        Iz = Bz / Bslope_TperA[chZ]

        logger.info(f'Setting {Bmag=}, {theta=} {phi=} in radians')
        logger.info(f"Setting {chX=} to {Ix}")
        logger.info(f"Setting {chY=} to {Iy}")
        logger.info(f"Setting {chZ=} to {Iz}")
        self.setChanIout(chX, Ix)
        self.setChanIout(chY, Iy)
        self.setChanIout(chZ, Iz)
        return Ix, Iy, Iz

    def setBinDegrees(self, Bmag=50e-6, theta=0, phi=0):
        logger.info(f'Setting {Bmag=},  {theta=} {phi=} in degrees')
        return self.setB(Bmag=Bmag, theta=theta/180*np.pi, phi=phi/180*np.pi)

# TSDB logger setting
tsdb_client = tsdb.Client('influx', 'http://lumus.physics.wm.edu:8428', database='qolab')
tsdb_ingester = tsdb.Ingester(tsdb_client, batch=11, measurement_prefix='VAMPIRE.VCSEL')

# creating Apparatus with all instruments to be logged
app_nickname = 'magnetometer'
apparatus = Apparatus(tsdb_ingester=tsdb_ingester, device_nickname=app_nickname,)
apparatus.config['Device type'] = 'QOL VAMPIRE VCSEL magnetometer'
apparatus.config['Device model'] = 'v0.1'
apparatus.config['FnamePrefix'] = 'magnetometer_eit'
# apparatus.config['SavePath'] = '/mnt/qol_grp_data/data.VAMPIRE.VCSEL'
apparatus.config['SavePath'] = './data'


logger.info("Accessing hardware")
rm = pyvisa.ResourceManager()
instr=rm.open_resource('USB0::10893::4354::MY61001869::0::INSTR')
Bfield = BfieldDriver(instr, device_nickname='.'.join([app_nickname, 'b_field_driver']), tsdb_ingester=tsdb_ingester)
# ps = KeysightE3612A(instr, device_nickname='.'.join([app_nickname, 'coil_driver']), tsdb_ingester=tsdb_ingester)
ps = Bfield # alias
# set safety current limit
ps.setChanIlimit(1, 0.1) # max current in Amps
ps.setChanIlimit(2, 0.1) # max current in Amps
ps.setChanIlimit(3, 0.1) # max current in Amps

if platform.system() == 'Linux':
    rfgen_port='/dev/ttyUSB0'
else:
    rfgen_port='COM4'
rfgen = QOL_LMX2487(port=rfgen_port, speed=115200, timeout=1, device_nickname='.'.join([app_nickname, 'rfgen']), tsdb_ingester=tsdb_ingester)


daq = LabJackUE9(device_nickname='.'.join([app_nickname, 'daq']), tsdb_ingester=tsdb_ingester)

logger.info("Adding instruments to apparatus")
apparatus.instruments={}
ai = apparatus.instruments
ai['rfgen'] = rfgen
ai['daq']   = daq
ai['b_field_driver'] = Bfield


logger.info('Setting magnetic field coils currents')
# ps.setChanIout_mA(1, 70)
# ps.setChanIout_mA(2, 0)
# ps.setChanIout_mA(3, 0)
logger.info('Done setting magnetic field coils currents')

def eitSweep(central_frequency, frequency_span, Np, Nsweeps=1):
    frList = np.linspace(central_frequency-frequency_span/2, central_frequency+frequency_span/2, Np)
    trFreq=Trace('Frequency')
    trFreq.config['unit']='Hz'
    trTransmission=Trace('Transmission')
    trTransmission.config['unit']='Arb. Unit'
    trLockin=Trace('Lockin')
    trLockin.config['unit']='V'

    trEIT = TraceSetSameX('EIT')

    for sw in pbar(range(1,Nsweeps+1), desc="Sweep"):
        for fr in pbar(frList, desc="Freq Scan"):
            rfgen.setFreqFixed(float(fr))
            time.sleep(dwellTime)
            transmission = daq.getAIN(0)
            lockin = daq.getAIN(1)

            trFreq.addPoint(fr)
            trTransmission.addPoint(transmission)
            trLockin.addPoint(lockin)

    # trFreq.values = trFreq.values - central_frequency
    trEIT.addTraceX(trFreq)
    trEIT.addTrace(trTransmission)
    trEIT.addTrace(trLockin)

    trEIT.config['tags']['apparatus']=apparatus.getConfig()

    return trEIT

def getCurrentCalibrationData(ch=1):
    curList = [20, 30, 40, 50, 60, 70, 80, 90, 100, 110]
    ps.setChanIout_mA(1, 0)
    ps.setChanIout_mA(2, 0)
    ps.setChanIout_mA(3, 0)
    for current in curList:
        logger.info(f'Preparing data for {current=} in {ch=} ready')
        ps.setChanIout_mA(ch, current)
        trEIT = eitSweep(central_frequency, frequency_span, Np, Nsweeps=Nsweeps)
        trEIT.plot()
        fn = apparatus.getNextDataFile()
        logger.info(f'Data ready for {current=} in {ch=}')
        logger.info(f'Data saved to {fn=}')
        trEIT.save(fn)

def calibrateCoilsCurrent():
    Np=1000
    Nsweeps=5
    getCurrentCalibrationData(ch=1)
    getCurrentCalibrationData(ch=2)
    getCurrentCalibrationData(ch=3)

def setBandTakeTrace(Bmag=50e-6, theta=0, phi=0, central_frequency=6.83468e9, frequency_span=2500e3, Np=100, Nsweeps=1):
    """ theta and phi assumed to be in degrees """
    Bfield.setBinDegrees(Bmag=Bmag, theta=theta, phi=phi)
    trEIT = eitSweep(central_frequency, frequency_span, Np, Nsweeps=Nsweeps)
    return trEIT


def rotateBandGetEITtrace():
    Np=500
    Nsweeps=5
    Bmag=50e-6 # earth magnetic field in Tesla (0.5 G)
    phiStep = 10
    thetaStep = 10
    phiSet = range(0,90+phiStep, phiStep)
    thetaSet = range(0,90+thetaStep, thetaStep)
    for phi in phiSet:
        for theta in thetaSet:
            trEIT = setBandTakeTrace(Bmag=Bmag, theta=theta, phi=phi, central_frequency=central_frequency, frequency_span=frequency_span, Np=Np, Nsweeps=Nsweeps)
            plt.clf()
            trEIT.plot()
            plt.draw()
            plt.pause(0.1)
            fn = apparatus.getNextDataFile()
            logger.info(f'Data ready for {Bmag=} in {theta=} {phi=}')
            logger.info(f'Data saved to {fn=}')
            trEIT.save(fn)


central_frequency = 6.83468e9
frequency_span =  2500e3

dwellTime=0.1
Np=100
Nsweeps=1


tsdb_ingester.commit()



from qolab.hardware.basic import BasicInstrument
from qolab.hardware.rf_generator import RFGenerator
import serial
import re
import time


class QOL_LMX2487(RFGenerator):
    """
    QOL lab made RF generator based on TI LMX2487 chip.
    The communication with RF generator is done via nodeMCU controlling esp2866 chip.
    """
    def __init__(self, *args,  port='/dev/ttyUSB0', speed=115200, timeout=1, setToDefaults=False, **kwds):
        super().__init__(*args, **kwds)
        self.config['Device model'] = 'QOL made RF generator based on TI LMX2487 chip'
        self.port = port
        self.speed = speed
        self.timeout = timeout
        self.connection = serial.Serial( self.port, self.speed, timeout=self.timeout)
        self.log = []
        self.logCapacity = 10
        self._FreqFixed = None
        self.hopeFree = True
        self.hopeFreeFreqJump = 100e3 # we break laser lock if Freq change is larger
        self.dwellTime=0.1 # needed for hope free setling
        if setToDefaults:
            self.sendSerialCmd("set_lmx2487_board_to_default_state()")

    def add2log(self, text):
        self.log.append(text)
        while len(self.log) > self.logCapacity:
            self.log.pop(0)

    def log2str(self, interval=None):
        strOut = ""
        for e in self.log:
            strOut += e
        return strOut

    def sendSerialCmd(self, cmd):
        self.connection.write(bytes(cmd+'\r','ascii'))
        if "3.4" == serial.__version__:
            # older version style
            resp = self.connection.read_until(terminator=b'> ')
        else:
            # new style after 20180616
            resp = self.connection.read_until(expected=b'> ')
        resp = resp.decode('utf-8')
        self.add2log(resp)
        return resp

    @BasicInstrument.tsdb_append
    def setFreqFixed(self,freq):
        """ 
        Will do incremental hope free tuning to desired frequency,
        if self.hopeFree is True.
        RF generator itself is fine, but our laser looses lock without it.
        """

        finished = False
        while not finished:
            frNow = self.getFreqFixed()
            dF = freq-frNow
            if dF >= 0:
                dFSign = 1
            else:
                dFSign = -1
            dF = abs(dF)
            if dF <= self.hopeFreeFreqJump or not self.hopeFree:
                fr = freq
                finished = True
            else:
                fr = frNow + dFSign * self.hopeFreeFreqJump
            self._FreqFixed = fr
            cmd_str=f'setFreq({fr:.2f})'
            self.sendSerialCmd(cmd_str)
            if not finished:
                time.sleep(self.dwellTime)

    @BasicInstrument.tsdb_append
    def getFreqFixed(self):
        # talking to hardware is slow, so we use cached value if we can
        if self._FreqFixed is None:
            resp = self.sendSerialCmd('getFreq()')
            m = re.search('[0-9.]+', resp)
            if m is not None:
                self._FreqFixed = float(m.group())
        return self._FreqFixed


if __name__ == '__main__': 
    import platform
    if platform.system() == 'Linux':
        rfgen=QOL_LMX2487(port='/dev/ttyUSB0', speed=115200, timeout=1)
    else:
        rfgen=QOL_LMX2487(port='COM4', speed=115200, timeout=1)
    print("testing")
    print('------ Header start -------------')
    print(str.join('\n', rfgen.getHeader()))
    print('------ Header ends  -------------')



import pytest
from qolab.hardware.scpi import response2numStr

def test_separator_and_unit():
    assert response2numStr('TDIV 2.00E-08S', firstSeparator=' ', unit='Z') == ('TDIV', '2.00E-08S', None) # incorrect unit was specified
    assert response2numStr('TDIV 2.00E-08S', firstSeparator=' ', unit='S') == ('TDIV', '2.00E-08', 'S')
    assert response2numStr('TDIV,2.00E-08S', firstSeparator=',', unit='S') == ('TDIV', '2.00E-08', 'S')

def test_separator_and_empty_unit():
    assert response2numStr('TDIV,2.00E-08', firstSeparator=',', unit=None) == ('TDIV', '2.00E-08', None)
    assert response2numStr('TDIV,2.00E-08', firstSeparator=',', unit='')   == ('TDIV', '2.00E-08', '')
                              
def test_no_separator_with_unit():
    assert response2numStr('2.00E-08S', firstSeparator=None, unit='S') == (None, '2.00E-08', 'S')
    assert response2numStr('2.00E-08S', firstSeparator='', unit='S')   == (None, '2.00E-08', 'S')
                              
def test_no_separator_and_no_unit():
    assert response2numStr('2.00E-08', firstSeparator=None, unit=None) == (None, '2.00E-08', None)
    assert response2numStr('2.00E-08', firstSeparator='', unit='')   == (None, '2.00E-08', '')
                              
